# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""File uploader."""

import logging
import mmap
import os
import time as tm
from pathlib import Path
from typing import IO, Literal, Protocol, override

import requests
from requests import status_codes

from debusine.client import exceptions
from debusine.client.client_utils import requests_put_or_connection_error
from debusine.utils import parse_range_header
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT


class _SupportsReadAndSeek[T](Protocol):
    """A file-like object supporting the ``read`` and ``seek`` methods."""

    def read(self, n: int | None = None, /) -> T:
        """Read data."""

    def seek(
        self,
        pos: int,
        whence: Literal[0, 1, 2] = ...,  # noqa: F841
        /,
    ) -> None:
        """Change the stream position to the given offset."""


class UnexpectedStatusCodeError(Exception):
    """Raised when the server did not return the expected status code."""

    def __init__(self, response: requests.models.Response) -> None:
        """Initialize exception."""
        self._response = response

    @override
    def __str__(self) -> str:
        """Return human-readable error message."""
        request = self._response.request

        error = f"Client HTTP {request.method} {request.url} "

        content_range = request.headers.get("Content-Range", None)

        if content_range is not None:
            error += f"with Content-Range {content_range} "

        error += (
            f"Received unexpected status code: {self._response.status_code} "
            f"Response body:\n{self._response.text}"
        )

        return error


class FileUploadError(Exception):
    """File could not be uploaded."""


class FileUploader:
    """
    Class used to upload files to debusine.

    Before uploading a file requests the server (via Content-Range header)
    which part of the file needs to be uploaded (or if all the file is
    already uploaded).
    """

    # Retries if a file upload fails (e.g. chunks cannot be uploaded or
    # the server's response is a file conflict, etc.)
    _MAX_FILE_UPLOAD_RETRIES = 3

    # Retries if a chunk upload fails (e.g. unexpected response from
    # the server)
    _MAX_CHUNK_UPLOAD_RETRIES = 6

    # Seconds to wait before retrying to upload a chunk again
    # (e.g. if it failed because connection was lost, unexpected response
    # from the server)
    _CHUNK_UPLOAD_RETRY_WAIT_SECS = 5

    # Size of the chunks. If changing it you might need to change Nginx's
    # client_max_body_size as well
    _MAX_CHUNK_SIZE_BYTES = 50 * 1024 * 1024

    def __init__(
        self,
        token: str,
        scope: str,
        *,
        logger: logging.Logger,
        timeout: int | tuple[int, int] = DEFAULT_REQUEST_TIMEOUT,
    ) -> None:
        """Initialize object."""
        self._headers = {
            "Token": token,
            "X-Debusine-Scope": scope,
        }
        self._logger = logger
        self._timeout = timeout

    def _upload_chunk_sleep(self, seconds: float) -> None:  # pragma: no cover
        """Split out from ``_upload_chunk`` to ease testing."""
        tm.sleep(seconds)

    def _upload_chunk(
        self,
        *,
        url: str,
        start_position: int,
        end_position: int,
        file_size: int,
        file: _SupportsReadAndSeek[bytes],
    ) -> requests.models.Response:
        content_range = f"bytes {start_position}-{end_position}/{file_size}"
        attempt = 0

        while True:
            file.seek(start_position)
            try:
                r = requests_put_or_connection_error(
                    url,
                    headers={**self._headers, "Content-Range": content_range},
                    data=file,
                    timeout=self._timeout,
                )
                return r
            except exceptions.ClientConnectionError as exc:
                attempt += 1

                if attempt == self._MAX_CHUNK_UPLOAD_RETRIES:
                    raise exc

                self._upload_chunk_sleep(self._CHUNK_UPLOAD_RETRY_WAIT_SECS)

    @staticmethod
    def _parse_range_or_raise_exception(
        response: requests.models.Response, file_size: int
    ) -> dict[str, int]:
        try:
            parsed_range = parse_range_header(response.headers, file_size)
            if parsed_range is not None:
                return parsed_range
        except ValueError:
            pass

        raise RuntimeError(
            "Required range header not in response or invalid for: "
            f"{response.request.method} "
            f"{response.request.url} content-range header: "
            f"{response.headers.get('content-range', None)}"
        )

    def _upload_file(self, file: IO[bytes], file_size: int, url: str) -> bool:
        content_range = f"bytes */{file_size}"

        r = requests_put_or_connection_error(
            url,
            headers={**self._headers, "Content-Range": content_range},
        )

        if r.status_code not in (
            status_codes.codes.ok,
            status_codes.codes.partial_content,
        ):
            raise UnexpectedStatusCodeError(r)

        if r.status_code == status_codes.codes.ok:
            # The file is already uploaded
            return True

        parsed_range = self._parse_range_or_raise_exception(r, file_size)

        start_position = parsed_range["end"]

        while True:
            end_position = min(
                file_size - 1, start_position + self._MAX_CHUNK_SIZE_BYTES - 1
            )

            # Create filelike object of the right chunk for requests.put
            with mmap.mmap(
                file.fileno(), end_position + 1, prot=mmap.PROT_READ
            ) as file_mmaped:
                r = self._upload_chunk(
                    url=url,
                    start_position=start_position,
                    end_position=end_position,
                    file_size=file_size,
                    file=file_mmaped,
                )

            if r.status_code == requests.codes.created:
                return True
            elif r.status_code == requests.codes.ok:
                parsed_range = self._parse_range_or_raise_exception(
                    r, file_size
                )
                start_position = parsed_range["end"]
            elif r.status_code == requests.codes.conflict:
                return False
            elif 500 <= r.status_code < 600:
                continue
            else:
                raise UnexpectedStatusCodeError(r)

    def upload(self, file_path: Path, url: str) -> None:
        """Upload the file used in this instance (see __init__())."""
        self._logger.info("Uploading %s to %s", file_path, url)

        upload_attempt = 1
        success = False

        while upload_attempt <= self._MAX_FILE_UPLOAD_RETRIES:
            with open(file_path, "rb") as file:
                success = self._upload_file(
                    file, os.stat(file_path).st_size, url
                )
                if success:
                    break
                upload_attempt += 1

        if not success:
            raise FileUploadError(
                f"Error uploading {file_path} to {url} "
                f"({self._MAX_FILE_UPLOAD_RETRIES} retries)"
            )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to create users."""

import secrets
import string
from typing import Any, NoReturn, override

from django.contrib.auth import get_user_model
from django.core.exceptions import ValidationError
from django.core.management import CommandError, CommandParser
from django.db import IntegrityError, transaction

from debusine.django.management.debusine_base_command import DebusineBaseCommand


class Command(DebusineBaseCommand):
    """Command to create a user."""

    help = "Create a new user. Output generated password on the stdout"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument("username", help="Username")
        parser.add_argument("email", help="Email for the user")

    @staticmethod
    def _generate_password() -> str:
        characters = string.ascii_letters + string.digits + string.punctuation
        length = 16
        return "".join(secrets.choice(characters) for i in range(length))

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Create the user."""
        password = self._generate_password()
        with transaction.atomic():
            try:
                user = get_user_model().objects.create_user(
                    username=options["username"],
                    email=options["email"],
                    password=password,
                )
                user.full_clean()
            except ValidationError as exc:
                raise CommandError(
                    "Error creating user: " + "\n".join(exc.messages),
                    returncode=3,
                )
            except IntegrityError:
                raise CommandError(
                    "A user with this username or email already exists",
                    returncode=3,
                )
            user.save()

        self.stdout.write(password)
        raise SystemExit(0)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""qa workflow."""

from typing import override

from django.db.models import Q, QuerySet

from debusine.artifacts import SourcePackage
from debusine.artifacts.models import TaskTypes
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact, WorkRequest
from debusine.server.collections.lookup import artifact_ids
from debusine.server.workflows import workflow_utils
from debusine.server.workflows.models import (
    AutopkgtestWorkflowData,
    BlhcWorkflowData,
    DebDiffWorkflowData,
    LintianWorkflowData,
    PiupartsWorkflowData,
    QAFailOn,
    QAWorkflowData,
    QAWorkflowDynamicData,
    RegressionTrackingUpdateQAResults,
    ReverseDependenciesAutopkgtestWorkflowData,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.regression_tracking import (
    RegressionTrackingWorkflow,
)
from debusine.tasks.models import (
    BackendType,
    InputArtifact,
    InputArtifactMultiple,
    InputArtifactSingle,
    LintianFailOnSeverity,
    LookupMultiple,
    LookupSingle,
    OutputData,
    RegressionAnalysis,
    RegressionAnalysisStatus,
)


class QAWorkflow(
    RegressionTrackingWorkflow[QAWorkflowData, QAWorkflowDynamicData]
):
    """QA workflow."""

    TASK_NAME = "qa"

    @override
    def compute_dynamic_data(self) -> QAWorkflowDynamicData:
        """
        Compute dynamic data for this workflow.

        :subject: package name of ``source_artifact``
        :parameter_summary: package name_version
        :source_artifact_id: id of the source artifact
        :binary_artifacts_ids: ids of the binary artifacts or None
        :reference_source_artifact_id: id of the reference artifact or None
        """
        source_artifact = workflow_utils.source_package(self)
        source_data = SourcePackage.create_data(source_artifact.data)

        if self.data.reference_source_artifact is not None:
            reference_source_artifact = workflow_utils.source_package(
                self, configuration_key="reference_source_artifact"
            )
        else:
            reference_source_artifact = None

        binary_artifacts_results = self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )

        return QAWorkflowDynamicData(
            subject=source_data.name,
            parameter_summary=f"{source_data.name}_{source_data.version}",
            source_artifact_id=source_artifact.id,
            binary_artifacts_ids=artifact_ids(binary_artifacts_results),
            reference_source_artifact_id=(
                reference_source_artifact.id
                if reference_source_artifact is not None
                else None
            ),
        )

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return the list of input artifacts."""
        artifacts: list[InputArtifact] = [
            InputArtifactSingle(
                lookup=self.data.source_artifact,
                label="source_artifact",
                artifact_id=(
                    self.dynamic_data.source_artifact_id
                    if self.dynamic_data is not None
                    else None
                ),
            ),
            InputArtifactMultiple.from_ids(
                lookup=self.data.binary_artifacts,
                label="binary_artifacts",
                artifact_ids=(
                    self.dynamic_data.binary_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            ),
        ]

        if self.data.reference_source_artifact is not None:
            artifacts.append(
                InputArtifactSingle(
                    lookup=self.data.reference_source_artifact,
                    label="reference_source_artifact",
                    artifact_id=(
                        None
                        if self.dynamic_data is None
                        else self.dynamic_data.reference_source_artifact_id
                    ),
                )
            )

        return artifacts

    @override
    def populate(self) -> None:
        """Create work requests."""
        if (data_archs := self.data.architectures) is not None:
            architectures = set(data_archs)
        else:
            architectures = workflow_utils.get_available_architectures(
                self, vendor=self.data.vendor, codename=self.data.codename
            )

        if (
            data_archs_allowlist := self.data.architectures_allowlist
        ) is not None:
            architectures.intersection_update(data_archs_allowlist)

        if (
            data_archs_denylist := self.data.architectures_denylist
        ) is not None:
            architectures.difference_update(data_archs_denylist)

        effective_architectures = sorted(architectures)

        filtered_binary_artifacts = (
            workflow_utils.filter_artifact_lookup_by_arch(
                self, self.data.binary_artifacts, architectures | {"all"}
            )
        )

        allow_failure = self.data.fail_on in {
            QAFailOn.NEVER,
            QAFailOn.REGRESSION,
        }

        if self.data.enable_autopkgtest:
            self._populate_autopkgtest(
                prefix=self.data.prefix,
                reference_prefix=self.data.reference_prefix,
                source_artifact=self.data.source_artifact,
                binary_artifacts=filtered_binary_artifacts,
                qa_suite=self.data.qa_suite,
                reference_qa_results=self.data.reference_qa_results,
                enable_regression_tracking=self.data.enable_regression_tracking,
                update_qa_results=self.data.update_qa_results,
                vendor=self.data.vendor,
                codename=self.data.codename,
                backend=self.data.autopkgtest_backend,
                architectures=effective_architectures,
                arch_all_build_architecture=(
                    self.data.arch_all_build_architecture
                ),
                allow_failure=allow_failure,
            )

        if self.data.enable_reverse_dependencies_autopkgtest:
            # Checked by
            # QAWorkflowData.check_reverse_dependencies_autopkgtest_consistency.
            assert self.data.qa_suite is not None
            self._populate_reverse_dependencies_autopkgtest(
                prefix=self.data.prefix,
                reference_prefix=self.data.reference_prefix,
                source_artifact=self.data.source_artifact,
                binary_artifacts=filtered_binary_artifacts,
                qa_suite=self.data.qa_suite,
                reference_qa_results=self.data.reference_qa_results,
                enable_regression_tracking=self.data.enable_regression_tracking,
                update_qa_results=self.data.update_qa_results,
                vendor=self.data.vendor,
                codename=self.data.codename,
                backend=self.data.autopkgtest_backend,
                architectures=effective_architectures,
                arch_all_build_architecture=(
                    self.data.arch_all_build_architecture
                ),
                allow_failure=allow_failure,
            )

        if self.data.enable_lintian:
            self._populate_lintian(
                prefix=self.data.prefix,
                reference_prefix=self.data.reference_prefix,
                source_artifact=self.data.source_artifact,
                binary_artifacts=filtered_binary_artifacts,
                qa_suite=self.data.qa_suite,
                reference_qa_results=self.data.reference_qa_results,
                enable_regression_tracking=self.data.enable_regression_tracking,
                update_qa_results=self.data.update_qa_results,
                vendor=self.data.vendor,
                codename=self.data.codename,
                backend=self.data.lintian_backend,
                architectures=effective_architectures,
                arch_all_build_architecture=(
                    self.data.arch_all_build_architecture
                ),
                fail_on_severity=self.data.lintian_fail_on_severity,
                allow_failure=allow_failure,
            )

        if self.data.enable_piuparts:
            self._populate_piuparts(
                prefix=self.data.prefix,
                reference_prefix=self.data.reference_prefix,
                source_artifact=self.data.source_artifact,
                binary_artifacts=filtered_binary_artifacts,
                qa_suite=self.data.qa_suite,
                reference_qa_results=self.data.reference_qa_results,
                enable_regression_tracking=self.data.enable_regression_tracking,
                update_qa_results=self.data.update_qa_results,
                vendor=self.data.vendor,
                codename=self.data.codename,
                architectures=effective_architectures,
                backend=self.data.piuparts_backend,
                environment=self.data.piuparts_environment,
                allow_failure=allow_failure,
            )

        if self.data.enable_debdiff:
            # Checked by
            # QAWorkflowData.enable_debdiff_consistency.
            assert self.data.qa_suite is not None
            self._populate_debdiff(
                source_artifact=self.data.source_artifact,
                binary_artifacts=filtered_binary_artifacts,
                vendor=self.data.vendor,
                codename=self.data.codename,
                original=self.data.qa_suite,
                arch_all_build_architecture=(
                    self.data.arch_all_build_architecture
                ),
                allow_failure=allow_failure,
            )

        if self.data.enable_blhc:
            self._populate_blhc(
                prefix=self.data.prefix,
                reference_prefix=self.data.reference_prefix,
                source_artifact=self.data.source_artifact,
                binary_artifacts=filtered_binary_artifacts,
                qa_suite=self.data.qa_suite,
                reference_qa_results=self.data.reference_qa_results,
                enable_regression_tracking=self.data.enable_regression_tracking,
                update_qa_results=self.data.update_qa_results,
                vendor=self.data.vendor,
                codename=self.data.codename,
                arch_all_build_architecture=(
                    self.data.arch_all_build_architecture
                ),
                allow_failure=allow_failure,
            )

        if self.data.fail_on == QAFailOn.REGRESSION:
            self._populate_regression_analysis_callback()

    def _populate_autopkgtest(
        self,
        *,
        prefix: str,
        reference_prefix: str,
        source_artifact: LookupSingle,
        binary_artifacts: LookupMultiple,
        qa_suite: LookupSingle | None,
        reference_qa_results: LookupSingle | None,
        enable_regression_tracking: bool,
        update_qa_results: RegressionTrackingUpdateQAResults,
        vendor: str,
        codename: str,
        backend: BackendType,
        architectures: list[str],
        arch_all_build_architecture: str,
        allow_failure: bool,
    ) -> None:
        """Create work request for autopkgtest workflow."""
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="autopkgtest",
            task_data=AutopkgtestWorkflowData(
                prefix=prefix,
                reference_prefix=reference_prefix,
                source_artifact=source_artifact,
                binary_artifacts=binary_artifacts,
                qa_suite=qa_suite,
                reference_qa_results=reference_qa_results,
                enable_regression_tracking=enable_regression_tracking,
                update_qa_results=update_qa_results,
                vendor=vendor,
                codename=codename,
                backend=backend,
                architectures=architectures,
                arch_all_build_architecture=arch_all_build_architecture,
                extra_repositories=self.data.extra_repositories,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name="autopkgtest",
                step="autopkgtest",
                allow_failure=allow_failure,
            ),
        )
        # The autopkgtest workflow's children will have dependencies on the
        # work requests creating source_artifact and binary_artifacts, but the
        # autopkgtest workflow itself doesn't need that in order to populate
        # itself.
        self.orchestrate_child(wr)

    def _populate_reverse_dependencies_autopkgtest(
        self,
        *,
        prefix: str,
        reference_prefix: str,
        source_artifact: LookupSingle,
        binary_artifacts: LookupMultiple,
        qa_suite: LookupSingle,
        reference_qa_results: LookupSingle | None,
        enable_regression_tracking: bool,
        update_qa_results: RegressionTrackingUpdateQAResults,
        vendor: str,
        codename: str,
        backend: BackendType,
        architectures: list[str],
        arch_all_build_architecture: str,
        allow_failure: bool,
    ) -> None:
        """Create work request for reverse_dependencies_autopkgtest workflow."""
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="reverse_dependencies_autopkgtest",
            task_data=ReverseDependenciesAutopkgtestWorkflowData(
                prefix=prefix,
                reference_prefix=reference_prefix,
                source_artifact=source_artifact,
                binary_artifacts=binary_artifacts,
                qa_suite=qa_suite,
                reference_qa_results=reference_qa_results,
                enable_regression_tracking=enable_regression_tracking,
                update_qa_results=update_qa_results,
                vendor=vendor,
                codename=codename,
                backend=backend,
                architectures=architectures,
                arch_all_build_architecture=arch_all_build_architecture,
                extra_repositories=self.data.extra_repositories,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name="autopkgtests of reverse-dependencies",
                step="reverse-dependencies-autopkgtest",
                allow_failure=allow_failure,
            ),
        )
        # The reverse_dependencies_autopkgtest workflow's descendants will have
        # dependencies on the work requests creating source_artifact and
        # binary_artifacts, but the reverse_dependencies_autopkgtest workflow
        # itself doesn't need that in order to populate itself.
        self.orchestrate_child(wr)

    def _populate_lintian(
        self,
        *,
        prefix: str,
        reference_prefix: str,
        source_artifact: LookupSingle,
        binary_artifacts: LookupMultiple,
        qa_suite: LookupSingle | None,
        reference_qa_results: LookupSingle | None,
        enable_regression_tracking: bool,
        update_qa_results: RegressionTrackingUpdateQAResults,
        vendor: str,
        codename: str,
        backend: BackendType,
        architectures: list[str],
        arch_all_build_architecture: str,
        fail_on_severity: LintianFailOnSeverity,
        allow_failure: bool,
    ) -> None:
        """Create work request for lintian workflow."""
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="lintian",
            task_data=LintianWorkflowData(
                prefix=prefix,
                reference_prefix=reference_prefix,
                source_artifact=source_artifact,
                binary_artifacts=binary_artifacts,
                qa_suite=qa_suite,
                reference_qa_results=reference_qa_results,
                enable_regression_tracking=enable_regression_tracking,
                update_qa_results=update_qa_results,
                vendor=vendor,
                codename=codename,
                backend=backend,
                architectures=architectures,
                arch_all_build_architecture=arch_all_build_architecture,
                fail_on_severity=fail_on_severity,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name="lintian",
                step="lintian",
                allow_failure=allow_failure,
            ),
        )
        # The lintian workflow's descendants will have dependencies on the work
        # requests creating source_artifact and binary_artifacts, but the
        # lintian workflow itself doesn't need that in order to populate
        # itself.
        self.orchestrate_child(wr)

    def _populate_piuparts(
        self,
        *,
        prefix: str,
        reference_prefix: str,
        source_artifact: LookupSingle,
        binary_artifacts: LookupMultiple,
        qa_suite: LookupSingle | None,
        reference_qa_results: LookupSingle | None,
        enable_regression_tracking: bool,
        update_qa_results: RegressionTrackingUpdateQAResults,
        vendor: str,
        codename: str,
        architectures: list[str],
        backend: BackendType,
        environment: LookupSingle | None,
        allow_failure: bool,
    ) -> None:
        data = PiupartsWorkflowData(
            prefix=prefix,
            reference_prefix=reference_prefix,
            source_artifact=source_artifact,
            binary_artifacts=binary_artifacts,
            qa_suite=qa_suite,
            reference_qa_results=reference_qa_results,
            enable_regression_tracking=enable_regression_tracking,
            update_qa_results=update_qa_results,
            vendor=vendor,
            codename=codename,
            architectures=architectures,
            backend=backend,
            extra_repositories=self.data.extra_repositories,
        )
        if environment is not None:
            data.environment = environment
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="piuparts",
            task_data=data,
            workflow_data=WorkRequestWorkflowData(
                display_name="piuparts",
                step="piuparts",
                allow_failure=allow_failure,
            ),
        )
        # The piuparts workflow's descendants will have dependencies on the
        # work requests creating binary_artifacts, but the piuparts workflow
        # itself doesn't need that in order to populate itself.
        self.orchestrate_child(wr)

    def _populate_debdiff(
        self,
        *,
        vendor: str,
        codename: str,
        source_artifact: LookupSingle,
        binary_artifacts: LookupMultiple,
        original: LookupSingle,
        arch_all_build_architecture: str,
        allow_failure: bool,
    ) -> None:
        data = DebDiffWorkflowData(
            original=original,
            source_artifact=source_artifact,
            binary_artifacts=binary_artifacts,
            vendor=vendor,
            codename=codename,
            arch_all_build_architecture=arch_all_build_architecture,
        )

        wr, _ = self.work_request_ensure_child_workflow(
            task_name="debdiff",
            task_data=data,
            workflow_data=WorkRequestWorkflowData(
                display_name="DebDiff",
                step="debdiff",
                allow_failure=allow_failure,
            ),
        )

        # Do not mark as running: the status is BLOCKED because the
        # binary_artifacts do not exist yet and DebDiffWorkflow needs
        # the binary artifacts in order to create the tasks
        self.requires_artifact(wr, binary_artifacts)

        self.orchestrate_child(wr)

    def _populate_blhc(
        self,
        *,
        prefix: str,
        reference_prefix: str,
        source_artifact: LookupSingle,
        binary_artifacts: LookupMultiple,
        qa_suite: LookupSingle | None,
        reference_qa_results: LookupSingle | None,
        enable_regression_tracking: bool,
        update_qa_results: RegressionTrackingUpdateQAResults,
        vendor: str,
        codename: str,
        arch_all_build_architecture: str,
        allow_failure: bool,
    ) -> None:
        wr, _ = self.work_request_ensure_child_workflow(
            task_name="blhc",
            task_data=BlhcWorkflowData(
                prefix=prefix,
                reference_prefix=reference_prefix,
                source_artifact=source_artifact,
                binary_artifacts=binary_artifacts,
                qa_suite=qa_suite,
                reference_qa_results=reference_qa_results,
                enable_regression_tracking=enable_regression_tracking,
                update_qa_results=update_qa_results,
                vendor=vendor,
                codename=codename,
                arch_all_build_architecture=arch_all_build_architecture,
            ),
            workflow_data=WorkRequestWorkflowData(
                display_name="build log hardening check",
                step="blhc",
                allow_failure=allow_failure,
            ),
        )

        # The blhc workflow needs the binary artifacts to exist with
        # relations to package build logs in order to populate itself.
        self.requires_artifact(wr, binary_artifacts)

        self.orchestrate_child(wr)

    def _find_regression_analysis_sources(self) -> QuerySet[WorkRequest]:
        """Find descendants that contribute regression analysis results."""
        return WorkRequest.objects.filter(
            Q(task_type=TaskTypes.WORKFLOW)
            & (
                # Most of the relevant tasks are children of this work
                # request.
                Q(parent=self.work_request)
                # Reverse-dependency autopkgtests are grandchildren of this
                # work request.
                | Q(parent__parent=self.work_request)
            )
        )

    def _populate_regression_analysis_callback(self) -> None:
        """Create a final callback to collate regression analysis results."""
        wr, _ = self.work_request_ensure_child_internal(
            task_name="workflow",
            workflow_data=WorkRequestWorkflowData(
                allow_failure=False,
                allow_dependency_failures=True,
                display_name="Regression analysis",
                step="regression-analysis",
                visible=True,
            ),
            unblock_strategy=WorkRequest.UnblockStrategy.INCREMENTAL,
        )
        for callback in WorkRequest.objects.workflow_callbacks().filter(
            workflow_data_json__step__startswith="regression-analysis",
            parent__in=self._find_regression_analysis_sources(),
        ):
            # Each individual regression analysis callback is a descendant
            # of this work request, and therefore must be in the same
            # workflow as the final callback.
            wr.add_dependency(callback, skip_same_workflow_check=True)

    @staticmethod
    def _summarize_status(
        statuses: set[RegressionAnalysisStatus],
    ) -> RegressionAnalysisStatus:
        if RegressionAnalysisStatus.REGRESSION in statuses:
            return RegressionAnalysisStatus.REGRESSION
        elif RegressionAnalysisStatus.ERROR in statuses:
            # Prefer REGRESSION if both exist, since it's probably more
            # informative: ERROR might be a temporary problem with a single
            # task.
            return RegressionAnalysisStatus.ERROR
        elif RegressionAnalysisStatus.IMPROVEMENT in statuses:
            return RegressionAnalysisStatus.IMPROVEMENT
        elif RegressionAnalysisStatus.STABLE in statuses:
            return RegressionAnalysisStatus.STABLE
        else:
            return RegressionAnalysisStatus.NO_RESULT

    def callback_regression_analysis(self) -> bool:
        """Collate regression analysis results."""
        assert self.dynamic_data is not None

        output_data = self.work_request.output_data or OutputData()
        output_data.regression_analysis = {}
        statuses: set[RegressionAnalysisStatus] = set()
        for (
            task_name,
            subject,
            analysis_by_architecture,
        ) in self._find_regression_analysis_sources().values_list(
            "task_name",
            "dynamic_task_data__subject",
            "output_data_json__regression_analysis",
        ):
            for key, analysis in (analysis_by_architecture or {}).items():
                # If we have output data, then dynamic task data must have
                # been populated, and all our sub-workflows have a subject.
                assert subject is not None
                output_data.regression_analysis[
                    f"{task_name}:{subject}:{key}"
                ] = RegressionAnalysis.model_validate(analysis)
                statuses.add(analysis["status"])

        # Top-level summary.
        overall_status = self._summarize_status(statuses)
        output_data.regression_analysis[""] = RegressionAnalysis(
            original_source_version=(
                SourcePackage.create_data(
                    Artifact.objects.get(
                        id=self.dynamic_data.reference_source_artifact_id
                    ).data
                ).version
                if self.dynamic_data.reference_source_artifact_id is not None
                else None
            ),
            new_source_version=(
                SourcePackage.create_data(
                    Artifact.objects.get(
                        id=self.dynamic_data.source_artifact_id
                    ).data
                ).version
                if self.dynamic_data.source_artifact_id is not None
                else None
            ),
            status=overall_status,
        )

        self.work_request.output_data = output_data
        self.work_request.save()
        return overall_status not in {
            RegressionAnalysisStatus.REGRESSION,
            RegressionAnalysisStatus.ERROR,
        }



# Generated by Django 4.2.23 on 2025-12-12 13:57

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0036_rename_create_child_workspace'),
    ]

    operations = [
        migrations.AlterField(
            model_name='collection',
            name='full_history_retention_period',
            field=models.DurationField(
                blank=True,
                help_text='Period after removing an item from the collection during which the related artifact is not allowed to expire',
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name='collection',
            name='metadata_only_retention_period',
            field=models.DurationField(
                blank=True,
                help_text='Period after removing an item from the collection during which its metadata is kept in the history of the collection',
                null=True,
            ),
        ),
        migrations.AlterField(
            model_name='collection',
            name='retains_artifacts',
            field=models.CharField(
                choices=[
                    ('never', 'Never'),
                    ('workflow', 'While workflow is running'),
                    ('always', 'Always'),
                ],
                default='always',
                help_text='Controls if artifacts in this collection should be retained rather than being allowed to expire',
                max_length=8,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to delete tokens. Allows filtering."""

from typing import Any, override

from django.core.management import CommandError, CommandParser

from debusine.db.models import Token
from debusine.django.management.debusine_base_command import DebusineBaseCommand
from debusine.server.management.management_utils import Tokens


class Command(DebusineBaseCommand):
    """Command to delete tokens."""

    help = (
        "Delete existing tokens. By default it requests confirmation "
        "before deletion"
    )

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            '--yes', action='store_true', help='Skips confirmation of deletion'
        )
        parser.add_argument(
            '--force',
            action='store_true',
            help='Does not fail if trying to delete nonexistent tokens',
        )
        parser.add_argument('--token', help='Deletes only this token')
        parser.add_argument(
            '--username', help='Deletes tokens owned by USERNAME'
        )

    @override
    def handle(self, *args: Any, **options: Any) -> None:
        """Delete (if selected and confirmed) tokens."""
        tokens = Token.objects.get_tokens(
            username=options["username"],
            key=options["token"],
            include_expired=True,
        )

        if not tokens:
            error_message = 'There are no tokens to be deleted'
            if options['force']:
                self.stdout.write(error_message)
                return
            else:
                raise CommandError(error_message, returncode=3)

        Tokens(yaml=False).print(tokens, self.stdout)

        deletion_confirmed = False
        deletion_forced = options['yes']
        if not deletion_forced:
            deletion_answer = input("Would you like to delete them? [yN] ")
            deletion_confirmed = deletion_answer.strip() in ('y', 'Y')

        if deletion_forced or deletion_confirmed:
            tokens.delete()



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""debusine-admin command to delete expired objects."""

import datetime as dt
import io
from typing import Any, NoReturn, override

import pgtransaction
from django.core.management import CommandParser
from django.db import transaction
from django.db.models import F, ProtectedError, Q, QuerySet
from django.template.defaultfilters import pluralize
from django.utils import timezone
from django_cte import CTE, with_cte

from debusine.db.models import (
    Artifact,
    ArtifactRelation,
    CollectionItem,
    File,
    FileInArtifact,
    FileInStore,
    Group,
    Token,
    WorkRequest,
    Workspace,
    work_requests,
)
from debusine.db.models.workspaces import DeleteWorkspaces
from debusine.django.management.debusine_base_command import DebusineBaseCommand

RETRY_ATTEMPTS = 3


class Command(DebusineBaseCommand):
    """Command to expire artifacts."""

    help = "Delete expired artifacts if no other artifact depends on them"

    @override
    def add_arguments(self, parser: CommandParser) -> None:
        parser.add_argument(
            "--dry-run",
            help="Trial run: do not delete anything",
            action="store_true",
        )

    @override
    def handle(self, *args: Any, **options: Any) -> NoReturn:
        """Delete expired artifacts."""
        with DeleteOperation(
            out=self.stdout,
            err=self.stderr,
            dry_run=options["dry_run"],
            verbosity=options["verbosity"],
        ) as operation:
            delete_expired_workrequests = DeleteExpiredWorkRequests(operation)
            delete_expired_workrequests.run()

            delete_expired_artifacts = DeleteExpiredArtifacts(operation)
            delete_expired_artifacts.run()

            delete_expired_ephemeral_groups = DeleteExpiredEphemeralGroups(
                operation
            )
            delete_expired_ephemeral_groups.run()

            delete_expired_workspaces = DeleteExpiredWorkspaces(operation)
            delete_expired_workspaces.run()

            delete_expired_tokens = DeleteExpiredTokens(operation)
            delete_expired_tokens.run()

        raise SystemExit(0)


class DeleteOperation:
    """Coordinate multiple delete operations."""

    def __init__(
        self,
        *,
        out: io.TextIOBase,
        err: io.TextIOBase,
        dry_run: bool = True,
        verbosity: int = 0,
    ) -> None:
        """
        Initialize object.

        :param out: to write information of deleted artifacts.
        :param err: to write errors that might happen during the deletion.
        :param dry_run: If True, no changes will be made to the database or
            files; instead the method will only simulate the execution
            of the actions that would have been taken.
        :param verbosity: If True, the method will provide detailed information
            about the execution progress and results.
        """
        self._out: io.TextIOBase = out
        self._err: io.TextIOBase = err

        self._verbosity: int = verbosity
        self.dry_run: bool = dry_run
        # datetime where all delete operations started
        self.initial_time: dt.datetime = (
            timezone.now()
        )  # time used for all the checks

    def __enter__(self) -> "DeleteOperation":
        """Context manager start."""
        if self.dry_run:
            self.verbose("dry-run mode: no changes will be made\n")
        return self

    def __exit__(self, *args: Any) -> None:
        """Context manager end."""
        if self.dry_run:
            self.verbose("dry-run mode: no changes were made\n")
        return None

    def verbose(self, message: str) -> None:
        """Print a verbose message."""
        if self._verbosity > 1:
            self._out.write(message)
            self._out.flush()


class DeleteExpiredArtifacts:
    """
    Delete expired artifacts.

    Delete expired artifacts that don't have a dependent artifact (direct or
    indirect via other artifacts).

    Delete other artifact/files associated models.

    Implemented using Mark and Sweep.
    """

    def __init__(self, operation: DeleteOperation) -> None:
        """
        Initialize object.

        :param operation: class used for output
        """
        self.operation = operation

    def run(self) -> None:
        """Delete the expired artifacts (with dry_run or verbosity levels)."""
        self._expire_collection_artifacts()
        self._expire_collection_metadata()
        self._expire_artifacts()
        self._delete_files_from_stores()

    @pgtransaction.atomic(
        isolation_level=pgtransaction.REPEATABLE_READ, retry=RETRY_ATTEMPTS
    )
    def _expire_collection_artifacts(self) -> None:
        """Drop artifacts from collections, retaining metadata."""
        CollectionItem.objects.drop_full_history(self.operation.initial_time)
        if self.operation.dry_run:
            transaction.set_rollback(True)

    @pgtransaction.atomic(
        isolation_level=pgtransaction.REPEATABLE_READ, retry=RETRY_ATTEMPTS
    )
    def _expire_collection_metadata(self) -> None:
        """Delete old collection items."""
        CollectionItem.objects.drop_metadata(self.operation.initial_time)
        if self.operation.dry_run:
            transaction.set_rollback(True)

    @pgtransaction.atomic(
        isolation_level=pgtransaction.REPEATABLE_READ, retry=RETRY_ATTEMPTS
    )
    def _expire_artifacts(self) -> None:
        """Drop old artifacts."""
        self._sweep(self._mark_to_keep())
        if self.operation.dry_run:
            transaction.set_rollback(True)

    def _mark_to_keep(
        self, *, keep_extra: QuerySet[Artifact] | None = None
    ) -> QuerySet[Artifact]:
        """
        Return a queryset of artifacts that must be kept.

        These artifacts cannot be deleted: they are not expired, or they
        would be expired but there is a relation that means they must be
        kept.

        The following artifact relations cause ``expired_artifact`` to be
        kept:

        * artifact=non_expired_artifact, target=expired_artifact, type=(any)
        * artifact=expired_artifact, target=non_expired_artifact, type=EXTENDS

        :param keep_extra: Keep these additional artifacts and anything
          marked as a result of them.
        """

        def keep_cte(cte: CTE) -> QuerySet[Artifact, Any]:
            # These artifacts are to be kept in their own right, either
            # because their expiration delays have not passed or because
            # they are part of a relevant collection.
            initial = (
                Artifact.objects.not_expired(self.operation.initial_time)
                | Artifact.objects.part_of_collection_with_retains_artifacts()
            )
            if keep_extra is not None:
                initial |= keep_extra

            # These relations cause the artifacts identified by kept_id to
            # be kept even if they would normally be expired, if keeper_id
            # is in the set of artifacts to keep.
            normal_relations = ArtifactRelation.objects.values(
                keeper_id=F("artifact"), kept_id=F("target")
            )
            extends_relations = ArtifactRelation.objects.filter(
                type=ArtifactRelation.Relations.EXTENDS
            ).values(keeper_id=F("target"), kept_id=F("artifact"))
            relations = CTE(
                normal_relations.union(extends_relations), name="relations"
            )

            # Follow relations recursively to establish the full set of
            # artifacts to keep.
            recursive = initial.values(keep_id=F("id")).union(
                with_cte(
                    relations,
                    select=cte.join(
                        relations.queryset(), keeper_id=cte.col.keep_id
                    ),
                ).values(keep_id=F("kept_id"))
            )
            assert isinstance(recursive, QuerySet)
            return recursive

        cte = CTE.recursive(keep_cte, name="keep")
        keep = with_cte(cte, select=cte.join(Artifact, id=cte.col.keep_id))
        assert isinstance(keep, QuerySet)
        return keep

    def _delete_artifact(self, artifact: Artifact) -> None:
        """
        Delete the specified Artifact and related models.

        Delete ArtifactRelation with target or artifact referencing artifact and
        the associated FileInArtifact files for this artifact.
        """
        if self.operation.dry_run:
            return

        for collection_item in artifact.collection_items.all():
            collection_item.parent_collection.manager.remove_item(
                collection_item
            )
        artifact.collection_items.update(artifact=None)

        ArtifactRelation.objects.filter(
            Q(target=artifact) | Q(artifact=artifact)
        ).delete()

        for file_in_artifact in FileInArtifact.objects.filter(
            artifact=artifact
        ):
            if hasattr(file_in_artifact, "fileupload"):
                file_in_artifact.fileupload.delete()
            file_in_artifact.delete()

        artifact.delete()

    @pgtransaction.atomic(
        isolation_level=pgtransaction.REPEATABLE_READ, retry=RETRY_ATTEMPTS
    )
    def _delete_files_from_stores(self) -> None:
        """
        Delete the files from stores.

        Files are not deleted at the same time as the artifacts. They need
        to be deleted after the transaction and only if no other FileInArtifact
        has the file.
        """
        orphaned_files = list(
            File.objects.filter(
                id__in=(
                    FileInStore.objects.values("file").difference(
                        FileInArtifact.objects.values("file")
                    )
                )
                # A SELECT ... FOR UPDATE lock prevents new references to
                # this file from being created while the lock is held.
                # There are no relations between files, so we don't need to
                # recompute anything if we need to skip some of these files
                # due to locks.
            ).select_for_update(skip_locked=True)
        )
        total = len(orphaned_files)
        if total:
            self.operation.verbose(
                f"Deleting {total} file{pluralize(total)} from the store\n"
            )

        for fileobj in orphaned_files:
            with transaction.atomic():
                # This is an orphaned file. Delete the file and
                # its dependencies
                for file_in_store in FileInStore.objects.filter(file=fileobj):
                    store = file_in_store.store.get_backend_object()
                    store.remove_file(fileobj)

                    try:
                        fileobj.delete()
                    except ProtectedError:
                        # Perhaps it's not possible to delete the file yet:
                        # it might be in another store
                        pass

    def _sweep(self, keep: QuerySet[Artifact]) -> None:
        """
        Delete artifacts that are not in ``keep``.

        This must be called within a transaction using the REPEATABLE READ
        isolation level.
        """
        number_of_artifacts_deleted = 0
        # Conceptually we should just use .exclude(id__in=keep) to find
        # expired artifacts instead, but that seems to cause PostgreSQL to
        # use a terrible plan that takes a very long time.  Pulling all the
        # IDs into Python is less elegant and will eventually be a scaling
        # problem, but with about a million IDs it only takes about four
        # seconds.
        keep_ids = set(keep.values_list("id", flat=True))
        expired = (
            Artifact.objects.expired(self.operation.initial_time)
            .exclude(id__in=keep_ids)
            .order_by("id")
        )
        expired_ids = {artifact.id for artifact in expired}
        expired_and_locked = expired.select_for_update(skip_locked=True)
        # We want this for safety checks later, but this also has the effect
        # of evaluating the queryset and therefore taking the lock.
        expired_and_locked_ids = {
            artifact.id for artifact in expired_and_locked
        }

        # Some expired artifacts may have been skipped due to a conflicting
        # lock, and some of those might be part of a relation chain that
        # should cause more artifacts to be kept.  Now that we have an upper
        # bound on the set of artifacts to expire, re-run the mark phase,
        # additionally keeping artifacts that we can't expire due to locks.
        # We're now within a FOR UPDATE lock on all the rows to be deleted,
        # so no new relations involving any of those rows can be added
        # concurrently (because that would require a FOR KEY SHARE lock on
        # the referenced rows, which conflicts with FOR UPDATE).
        unlocked_ids = expired_ids - expired_and_locked_ids
        if unlocked_ids:
            new_keep_ids = self._mark_to_keep(
                keep_extra=Artifact.objects.filter(id__in=unlocked_ids)
            )
            new_expired = (
                Artifact.objects.expired(self.operation.initial_time)
                .exclude(id__in=new_keep_ids)
                .order_by("id")
            )
            new_expired_ids = {artifact.id for artifact in new_expired}
            # Since we use the REPEATABLE READ isolation level,
            # nonrepeatable reads and (in PostgreSQL) phantom reads are
            # impossible, so _mark_to_keep must keep at least all the
            # rows it returned the first time round.
            if not (new_expired_ids <= expired_and_locked_ids):
                raise AssertionError(
                    "Second expired artifacts mark phase wants to expire"
                    " some artifacts that were not expired in the first"
                    " phase, despite REPEATABLE READ!"
                )
            expired_and_locked = new_expired

        for artifact_expired in expired_and_locked:
            artifact_id = artifact_expired.id
            number_of_artifacts_deleted += 1
            self._delete_artifact(artifact_expired)

            self.operation.verbose(f"Deleted artifact {artifact_id}\n")

        if number_of_artifacts_deleted == 0:
            self.operation.verbose("There were no expired artifacts\n")


class DeleteExpiredWorkRequests:
    """
    Delete expired work requests.

    Delete work requests that don't have unexpired child work requests.

    Delete also the work requests' internal collection if present.
    """

    to_delete: work_requests.DeleteUnused

    def __init__(self, operation: DeleteOperation) -> None:
        """
        Initialize object.

        :param operation: class used for output
        """
        self.operation = operation

    def perform_deletions(self) -> None:
        """Delete objects found in self.work_requests and self.collections."""
        self.to_delete.perform_deletions()

    @pgtransaction.atomic(
        isolation_level=pgtransaction.REPEATABLE_READ, retry=RETRY_ATTEMPTS
    )
    def run(self) -> None:
        """Find and optionally delete the expired work requests."""
        self.to_delete = work_requests.DeleteUnused(
            WorkRequest.objects.expired(at=self.operation.initial_time)
        )
        if self.to_delete.work_requests:
            self.operation.verbose(
                f"Deleting {len(self.to_delete.work_requests)}"
                " expired work requests"
                f" and {len(self.to_delete.collections)}"
                " expired collections\n"
            )

            # FIXME: do we enumerate each deleted work request to stdout?

            if not self.operation.dry_run:
                self.perform_deletions()
        else:
            self.operation.verbose("There were no expired work requests\n")


class DeleteExpiredEphemeralGroups:
    """Delete expired exphemeral groups."""

    groups: QuerySet[Group]

    def __init__(self, operation: DeleteOperation) -> None:
        """
        Initialize object.

        :param operation: class used for output
        """
        self.operation = operation

    def perform_deletions(self) -> None:
        """Delete previously found unused ephemeral groups."""
        self.groups.delete()

    @pgtransaction.atomic(
        isolation_level=pgtransaction.REPEATABLE_READ, retry=RETRY_ATTEMPTS
    )
    def run(self) -> None:
        """Find and optionally delete the expired work requests."""
        self.groups = Group.objects.unused_ephemeral()
        if count := len(self.groups):
            self.operation.verbose(
                f"Deleting {count} unused ephemeral groups\n"
            )
            if not self.operation.dry_run:
                self.perform_deletions()
        else:
            self.operation.verbose("There were no unused ephemeral groups\n")


class DeleteExpiredWorkspaces:
    """
    Delete expired workspaces.

    Delete workspaces that have expiration_delay set, ``expiration_delay`` days
    after ``max(created_at, last_task_completion_time)``
    """

    #: Work requests to delete
    to_delete: DeleteWorkspaces

    def __init__(self, operation: DeleteOperation) -> None:
        """
        Initialize object.

        :param operation: class used for output
        """
        self.operation = operation

    def perform_deletions(self) -> None:
        """Delete objects found in self.work_requests and self.collections."""
        self.to_delete.perform_deletions()

    @pgtransaction.atomic(
        isolation_level=pgtransaction.REPEATABLE_READ, retry=RETRY_ATTEMPTS
    )
    def run(self) -> None:
        """Find and optionally delete the expired work requests."""
        self.to_delete = DeleteWorkspaces(
            Workspace.objects.exclude(expiration_delay__isnull=True)
            .with_expiration_time()
            .filter(expiration_time__lt=self.operation.initial_time)
        )
        if self.to_delete.workspaces:
            self.operation.verbose(
                f"Deleting {len(self.to_delete.workspaces)}"
                " expired workspaces\n"
            )
            if not self.operation.dry_run:
                self.perform_deletions()
        else:
            self.operation.verbose("There were no expired workspaces\n")


class DeleteExpiredTokens:
    """Delete expired tokens."""

    tokens: QuerySet[Token]

    def __init__(self, operation: DeleteOperation) -> None:
        """
        Initialize object.

        :param operation: class used for output
        """
        self.operation = operation

    def perform_deletions(self) -> None:
        """Delete previously-found expired tokens."""
        self.tokens.delete()

    @pgtransaction.atomic(
        isolation_level=pgtransaction.REPEATABLE_READ, retry=RETRY_ATTEMPTS
    )
    def run(self) -> None:
        """Find and optionally delete expired tokens."""
        self.tokens = (
            Token.objects.expired(at=self.operation.initial_time)
            # Expiring tokens linked to workers would violate database
            # constraints.  The provisioner should clean up the most likely
            # case of this (a worker that never finished activating), after
            # which we can expire such tokens.
            .exclude(worker__isnull=False)
            .exclude(activating_worker__isnull=False)
        )
        if count := len(self.tokens):
            self.operation.verbose(f"Deleting {count} expired tokens\n")
            if not self.operation.dry_run:
                self.perform_deletions()
        else:
            self.operation.verbose("There were no expired tokens\n")



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for the management command delete_channel_notification."""

from typing import override

from django.core.management import CommandError
from django.test import TestCase

from debusine.db.models import NotificationChannel
from debusine.django.management.tests import call_command


class DeleteNotificationChannelTests(TestCase):
    """Tests for the delete_notification_channel command."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.email_data = {
            "from": "debusine@debusine.com",
            "to": ["user@example.com"],
        }

    def test_delete_notification_channel(self) -> None:
        """Delete notification channel."""
        name = "lts"
        NotificationChannel.objects.create(
            name=name,
            method=NotificationChannel.Methods.EMAIL,
            data=self.email_data,
        )

        stdout, stderr, exit_code = call_command(
            "delete_notification_channel", name
        )

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        self.assertEqual(exit_code, 0)

        self.assertEqual(NotificationChannel.objects.count(), 0)

    def test_delete_notification_channel_name_does_not_exist(self) -> None:
        """Delete notification channel fails: name does not exist."""
        name = "does-not-exist"
        expected_error = f'NotificationChannel "{name}" does not exist'

        with self.assertRaisesMessage(CommandError, expected_error) as exc:
            call_command("delete_notification_channel", name)

        self.assertEqual(exc.exception.returncode, 3)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for the DBTask input fields."""

import inspect
from types import SimpleNamespace
from typing import Any, ClassVar, override
from unittest import mock

from debusine.artifacts.models import ArtifactCategory, CollectionCategory
from debusine.db.models import Artifact, Collection, WorkRequest, Workspace
from debusine.db.models.lookups import LookupResult
from debusine.db.playground import scenarios
from debusine.db.tasks.inputs import (
    ArtifactInput,
    CollectionInSameWorkspace,
    CollectionInput,
    TaskConfigurationCollection,
)
from debusine.db.tasks.models import (
    DBArtifactInfo,
    DBInputCollectionSingle,
    DBInputLookupResult,
)
from debusine.db.tasks.resolver import DBFieldResolver
from debusine.tasks import TaskConfigError
from debusine.tasks.inputs.inputs import DataFieldInput
from debusine.tasks.models import BaseTaskData, LookupSingle
from debusine.test.django import TestCase
from debusine.utils import extract_generic_type_arguments


class InputTests[FieldType: DataFieldInput[Any, Any]](TestCase):
    """Test case that can work as a container for task input fields."""

    #: Field class to be tested, filled by __init_subclass__
    field_class: type[FieldType]

    #: Work request supporting the task
    work_request: ClassVar[WorkRequest]

    @override
    def __init_subclass__(cls, **kwargs: Any) -> None:
        """Fill lookup_type from typing information."""
        super().__init_subclass__(**kwargs)
        [cls.field_class] = extract_generic_type_arguments(cls, InputTests)

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.work_request = cls.playground.create_server_task()

    @override
    def setUp(self) -> None:
        super().setUp()
        # self.field is typed here, but set later by test methods
        self.field: FieldType

        # Task supporting the resolution
        self.task = self.work_request.get_task()

        # Field resolver
        self.resolver = DBFieldResolver(self.task)

        # Mock check_category
        self.check_category = self.enterContext(
            mock.patch(
                "debusine.tasks.inputs.inputs.BaseArtifactInput.check_category"
            )
        )

    def resolve(self) -> Any:
        """Call self.field.resolve and return its result."""
        return self.field.resolve(self.resolver)

    def assertIsDBArtifactInfo(
        self,
        value: Any,
        artifact: Artifact,
        lookup: LookupSingle | None = None,
        label: str | None = None,
    ) -> None:
        """Ensure the given value is a DBArtifactInfo matching the arguments."""
        lookup = lookup or self.resolver.get_field(self.field)
        assert lookup is not None
        label = label or self.field.field
        assert isinstance(value, DBArtifactInfo)
        self.assertEqual(
            value,
            DBArtifactInfo.from_artifact(artifact, lookup=lookup, label=label),
        )
        with self.assertNumQueries(0):
            self.assertEqual(value.artifact, artifact)

    def assertIsDBInputCollectionSingle(
        self,
        value: Any,
        collection: Collection,
        lookup: LookupSingle | None = None,
        label: str | None = None,
    ) -> None:
        """Ensure the value is a matching DBInputCollectionSingle."""
        lookup = lookup or self.resolver.get_field(self.field)
        assert lookup is not None
        label = label or self.field.field
        assert isinstance(value, DBInputCollectionSingle)
        self.assertEqual(
            value,
            DBInputCollectionSingle.from_collection(
                collection, lookup=lookup, label=label
            ),
        )
        with self.assertNumQueries(0):
            self.assertEqual(value.collection, collection)

    def assertIsDBInputLookupResult(
        self, value: Any, result: LookupResult, label: str | None = None
    ) -> None:
        """Ensure the value is a matching DBInputLookupResult."""
        label = label or self.field.field
        assert isinstance(value, DBInputLookupResult)
        self.assertEqual(
            value, DBInputLookupResult.from_lookup_result(result, label=label)
        )
        with self.assertNumQueries(0):
            self.assertEqual(value.result, result)

    def test_has_serializer(self) -> None:
        if not hasattr(self, "field_class") or inspect.isabstract(
            self.field_class
        ):
            return
        if not hasattr(self.field_class, "serializer"):
            self.fail(
                f"{self.field_class} has no serializer defined"
            )  # pragma: no cover


class ArtifactInputTests(InputTests[ArtifactInput]):
    """Tests for :py:class:`ArtifactInput`."""

    artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.artifact = cls.playground.create_minimal_source_package_artifact()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class(field="field")

    def test_none(self) -> None:
        lookup_single_artifact = self.enterContext(
            self.spy_object(self.resolver, "lookup_single_artifact")
        )
        self.task.data = SimpleNamespace(field=None)
        with self.assertRaises(AssertionError):
            self.resolve()
        lookup_single_artifact.assert_not_called()
        self.check_category.assert_not_called()

    def test_has_value(self) -> None:
        self.task.data = SimpleNamespace(field=self.artifact.pk)
        self.assertIsDBArtifactInfo(self.resolve(), self.artifact)

    def test_lookup_malformed(self) -> None:
        self.task.data = SimpleNamespace(field="invalid")
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"Lookup for field 'field' in {self.work_request.workspace}"
            rf" failed: 'invalid' does not specify a category",
        ):
            self.resolve()

    def test_lookup_fails(self) -> None:
        self.task.data = SimpleNamespace(field="invalid@debian:upload")
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"Lookup for field 'field' in {self.work_request.workspace}"
            rf" failed: 'invalid@debian:upload' does not exist or is hidden",
        ):
            self.resolve()

    def test_invalid_category(self) -> None:
        self.field = self.field_class(
            field="field", categories=[ArtifactCategory.TEST]
        )
        self.task.data = SimpleNamespace(field=self.artifact.pk)
        with self.assertRaisesRegex(
            TaskConfigError,
            "unexpected artifact category: 'debian:source-package'."
            r" Valid categories: \['debusine:test'\]",
        ):
            self.resolve()

    def test_get_artifact_ids(self) -> None:
        self.assertEqual(
            self.field.get_artifact_ids(
                DBArtifactInfo.from_artifact(
                    self.artifact, lookup=1, label="field"
                )
            ),
            [self.artifact.id],
        )


class CollectionInputTests(InputTests[CollectionInput]):
    """Tests for :py:class:`CollectionInput`."""

    collection: ClassVar[Collection]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.collection = cls.playground.create_collection(
            name="sid", category=CollectionCategory.SUITE
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class(field="field")

    def test_none(self) -> None:
        lookup_single = self.enterContext(
            self.spy("debusine.db.models.WorkRequest.lookup_single")
        )
        self.task.data = SimpleNamespace(field=None)
        with self.assertRaises(AssertionError):
            self.resolve()
        lookup_single.assert_not_called()

    def test_has_value(self) -> None:
        self.task.data = SimpleNamespace(field=self.collection.pk)
        self.assertIsDBInputCollectionSingle(self.resolve(), self.collection)

    def test_lookup_malformed(self) -> None:
        self.task.data = SimpleNamespace(field="invalid")
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"Lookup for field 'field' in {self.work_request.workspace}"
            rf" failed: 'invalid' does not specify a category",
        ):
            self.resolve()

    def test_lookup_fails(self) -> None:
        self.task.data = SimpleNamespace(field="invalid@debian:suite")
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"Lookup for field 'field' in {self.work_request.workspace}"
            rf" failed: 'invalid@debian:suite' does not exist or is hidden",
        ):
            self.resolve()

    def test_valid_category(self) -> None:
        self.field = self.field_class(
            field="field", category=CollectionCategory.SUITE
        )
        self.task.data = SimpleNamespace(field=self.collection.pk)
        self.assertIsDBInputCollectionSingle(self.resolve(), self.collection)

    def test_invalid_category(self) -> None:
        self.field = self.field_class(
            field="field", category=CollectionCategory.TEST
        )
        self.task.data = SimpleNamespace(field=self.collection.pk)
        with self.assertRaisesRegex(
            TaskConfigError,
            "unexpected collection category: 'debian:suite'."
            " Expected: 'debusine:test'",
        ):
            self.resolve()


class CollectionInSameWorkspaceTests(InputTests[CollectionInSameWorkspace]):
    """Tests for :py:class:`CollectionInSameWorkspace`."""

    other_workspace: Workspace
    collection: Collection
    other_collection: Collection

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.other_workspace = cls.playground.create_workspace(
            name="other", public=True
        )
        cls.collection = cls.playground.create_collection(
            name="sid",
            category=CollectionCategory.SUITE,
            workspace=cls.work_request.workspace,
        )
        cls.other_collection = cls.playground.create_collection(
            name="sid",
            category=CollectionCategory.SUITE,
            workspace=cls.other_workspace,
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class(field="field")

    def test_none(self) -> None:
        lookup_single = self.enterContext(
            self.spy("debusine.db.models.WorkRequest.lookup_single")
        )
        self.task.data = SimpleNamespace(field=None)
        with self.assertRaises(AssertionError):
            self.resolve()
        lookup_single.assert_not_called()

    def test_same_workspace(self) -> None:
        self.task.data = SimpleNamespace(
            field=f"sid@{CollectionCategory.SUITE}"
        )
        self.assertIsDBInputCollectionSingle(self.resolve(), self.collection)

    def test_different_workspace(self) -> None:
        lookup = f"//workspace=other/sid@{CollectionCategory.SUITE}"
        self.task.data = SimpleNamespace(field=lookup)
        with self.assertRaisesRegex(
            TaskConfigError,
            f"{lookup} is in workspace {self.other_workspace},"
            f" not {self.work_request.workspace}",
        ):
            self.resolve()


class TaskConfigurationCollectionTests(InputTests[TaskConfigurationCollection]):
    """Tests for :py:class:`TaskConfigurationCollection`."""

    scenario = scenarios.DefaultContext()

    @override
    def setUp(self) -> None:
        super().setUp()
        self.field = self.field_class(field="field")

    def test_model_default(self) -> None:
        data = BaseTaskData()
        self.assertEqual(
            data.task_configuration,
            BaseTaskData.DEFAULT_TASK_CONFIGURATION_LOOKUP,
        )

    def test_model_explicit_none(self) -> None:
        data = BaseTaskData(task_configuration=None)
        self.assertIsNone(data.task_configuration)

    def test_lookup_none(
        self,
    ) -> None:
        self.task.data = SimpleNamespace(field=None)
        result = self.resolve()
        self.assertIsNone(result)

    def test_failed_default(self) -> None:
        self.task.data = SimpleNamespace(
            field=BaseTaskData.DEFAULT_TASK_CONFIGURATION_LOOKUP
        )
        self.enterContext(
            mock.patch.object(
                self.work_request, "lookup_single", side_effect=KeyError()
            )
        )
        self.assertIsNone(self.resolve())

    def test_failed_custom(self) -> None:
        self.task.data = SimpleNamespace(field="custom")
        self.enterContext(
            mock.patch.object(
                self.work_request,
                "lookup_single",
                side_effect=LookupError("expected error"),
            )
        )
        with self.assertRaisesRegex(
            TaskConfigError,
            rf"Lookup for field 'field' in {self.work_request.workspace}"
            rf" failed: expected error",
        ):
            self.resolve()

    def test_success(self) -> None:
        self.task.data = SimpleNamespace(
            field=BaseTaskData.DEFAULT_TASK_CONFIGURATION_LOOKUP
        )
        self.assertIsDBInputCollectionSingle(
            self.resolve(),
            self.scenario.default_task_configuration_collection,
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Unit tests for task input field serializers."""

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianBinaryPackage,
)
from debusine.tasks.inputs.serializers import (
    ExtraRepositoriesSerializer,
    MultiInputListSerializer,
    OptionalPydanticSerializer,
    PydanticListSerializer,
    PydanticSerializer,
    ScalarSerializer,
    SuiteArchiveInputSerializer,
)
from debusine.tasks.models import (
    BaseTaskInputArtifact,
    ExtraExternalRepository,
    InputArtifactMultiple,
    InputCollectionSingle,
    LookupMultiple,
)
from debusine.test import TestCase


class SerializerTestCase(TestCase):
    """Base test case for serializer tests."""

    def make_multi(
        self,
        *artifacts: BaseTaskInputArtifact,
        label: str = "label",
        lookup: LookupMultiple = LookupMultiple(("lookup",)),
    ) -> InputArtifactMultiple:
        """Create an InputArtifactMultiple instance."""
        return InputArtifactMultiple(
            label=label, lookup=lookup, artifacts=list(artifacts)
        )


class ScalarSerializerTests(SerializerTestCase):
    """Tests for :py:class:`ScalarSerializer`."""

    def test_serialization(self) -> None:
        serializer = ScalarSerializer(int)
        self.assertEqual(serializer.serialize(3), 3)
        self.assertEqual(serializer.deserialize(3), 3)

    def test_invalid_value(self) -> None:
        serializer = ScalarSerializer(int)
        with self.assertRaisesRegex(
            ValueError, r"Value 'foo' is not an instance of <class 'int'>"
        ):
            serializer.deserialize("foo")


class PydanticSerializerTests(SerializerTestCase):
    """Tests for :py:class:`PydanticSerializer`."""

    sample = DebianBinaryPackage(
        srcpkg_name="unsigned",
        srcpkg_version="1.0",
        deb_fields={
            "Package": "hello",
            "Version": "1.0-1",
            "Architecture": "amd64",
        },
        deb_control_files=[],
    )

    def test_serialization(self) -> None:
        serializer = PydanticSerializer(DebianBinaryPackage)
        serialized = serializer.serialize(self.sample)
        assert isinstance(serialized, dict)
        deserialized = serializer.deserialize(serialized)
        self.assertEqual(deserialized, self.sample)


class OptionalPydanticSerializerTests(SerializerTestCase):
    """Tests for :py:class:`OptionalPydanticSerializer`."""

    sample = DebianBinaryPackage(
        srcpkg_name="unsigned",
        srcpkg_version="1.0",
        deb_fields={
            "Package": "hello",
            "Version": "1.0-1",
            "Architecture": "amd64",
        },
        deb_control_files=[],
    )

    def test_serialization(self) -> None:
        serializer = OptionalPydanticSerializer(DebianBinaryPackage)
        serialized = serializer.serialize(self.sample)
        assert isinstance(serialized, dict)
        deserialized = serializer.deserialize(serialized)
        self.assertEqual(deserialized, self.sample)

    def test_serialize_none(self) -> None:
        serializer = OptionalPydanticSerializer(DebianBinaryPackage)
        serialized = serializer.serialize(None)
        self.assertIsNone(serialized)
        deserialized = serializer.deserialize(None)
        self.assertIsNone(deserialized)


class PydanticListSerializerTests(SerializerTestCase):
    """Tests for :py:class:`PydanticSerializer`."""

    sample = DebianBinaryPackage(
        srcpkg_name="unsigned",
        srcpkg_version="1.0",
        deb_fields={
            "Package": "hello",
            "Version": "1.0-1",
            "Architecture": "amd64",
        },
        deb_control_files=[],
    )

    def test_serialization(self) -> None:
        serializer = PydanticListSerializer(DebianBinaryPackage)
        serialized = serializer.serialize([self.sample])
        assert isinstance(serialized, list)
        for idx, item in enumerate(serialized):
            with self.subTest(idx=idx):
                assert isinstance(item, dict)
        deserialized = serializer.deserialize(serialized)
        self.assertEqual(deserialized, [self.sample])

    def test_deserialize_invalid_type(self) -> None:
        serializer = PydanticListSerializer(DebianBinaryPackage)
        with self.assertRaisesRegex(
            ValueError, r"Value 'foo' is not an instance of <class 'list'>"
        ):
            serializer.deserialize("foo")


class MultiInputListSerializerTests(SerializerTestCase):
    """Tests for :py:class:`MultiInputListSerializer`."""

    sample = BaseTaskInputArtifact(
        artifact_id=1,
        category=ArtifactCategory.BINARY_PACKAGE,
        data=DebianBinaryPackage(
            srcpkg_name="unsigned",
            srcpkg_version="1.0",
            deb_fields={
                "Package": "hello",
                "Version": "1.0-1",
                "Architecture": "amd64",
            },
            deb_control_files=[],
        ),
    )

    def test_serialization_list(self) -> None:
        sample = [
            self.make_multi(self.sample, self.sample),
            self.make_multi(self.sample),
        ]
        serializer = MultiInputListSerializer()
        serialized = serializer.serialize(sample)
        assert isinstance(serialized, list)
        assert isinstance(serialized[0], dict)
        deserialized = serializer.deserialize(serialized)
        self.assertEqual(deserialized, sample)

    def test_deserialize_invalid_type(self) -> None:
        serializer = MultiInputListSerializer()
        with self.assertRaisesRegex(
            ValueError, r"Value 'foo' is not an instance of <class 'list'>"
        ):
            serializer.deserialize("foo")

    def test_deserialize_invalid_list_item_type(self) -> None:
        serializer = MultiInputListSerializer()
        with self.assertRaisesRegex(
            ValueError,
            "Input should be a valid dictionary"
            " or instance of InputArtifactMultiple",
        ):
            serializer.deserialize(["foo"])


class SuiteArchiveInputSerializerTests(SerializerTestCase):
    """Tests for :py:class:`SuiteArchiveInputSerializer`."""

    suite = InputCollectionSingle(
        label="field",
        lookup="lookup",
        collection_id=1,
        scope_name="scope",
        workspace_name="workspace",
        category=CollectionCategory.SUITE,
        name="sid",
        data={},
    )
    archive = InputCollectionSingle(
        label="field",
        lookup="lookup",
        collection_id=2,
        scope_name="scope",
        workspace_name="workspace",
        category=CollectionCategory.ARCHIVE,
        name="_",
        data={},
    )

    def test_serialization_both_set(self) -> None:
        sample = (self.suite, self.archive)
        serializer = SuiteArchiveInputSerializer()
        serialized = serializer.serialize(sample)
        assert isinstance(serialized, list)
        deserialized = serializer.deserialize(serialized)
        self.assertEqual(deserialized, sample)

    def test_serialization_archive_unset(self) -> None:
        sample = (self.suite, None)
        serializer = SuiteArchiveInputSerializer()
        serialized = serializer.serialize(sample)
        assert isinstance(serialized, list)
        deserialized = serializer.deserialize(serialized)
        self.assertEqual(deserialized, sample)

    def test_deserialize_invalid_type(self) -> None:
        serializer = SuiteArchiveInputSerializer()
        with self.assertRaisesRegex(
            ValueError, r"Value 'foo' is not an instance of <class 'list'>"
        ):
            serializer.deserialize("foo")

    def test_deserialize_invalid_length(self) -> None:
        serializer = SuiteArchiveInputSerializer()
        with self.assertRaisesRegex(
            ValueError, r"Value \[\] should be a list of exactly two items"
        ):
            serializer.deserialize([])


class ExtraRepositoriesSerializerTests(SerializerTestCase):
    """Tests for :py:class:`ExtraRepositoriesSerializer`."""

    sample = ExtraExternalRepository(
        url="https://example.org", suite="sid", components=["main"]
    )

    def test_serialization_list(self) -> None:
        serializer = ExtraRepositoriesSerializer()
        serialized = serializer.serialize([self.sample, self.sample])
        assert isinstance(serialized, list)
        deserialized = serializer.deserialize(serialized)
        self.assertEqual(deserialized, [self.sample, self.sample])

    def test_serialization_none(self) -> None:
        serializer = ExtraRepositoriesSerializer()
        serialized = serializer.serialize(None)
        self.assertIsNone(serialized)
        deserialized = serializer.deserialize(None)
        self.assertIsNone(deserialized)



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests for Debusine CLI archive token management commands."""

from functools import partial
from typing import override
from unittest import mock

from debusine.artifacts.models import CollectionCategory
from debusine.client.commands.archives.tokens import TokenCommand
from debusine.client.commands.base import Command, DebusineCommand
from debusine.client.commands.tests.base import BaseCliTests
from debusine.client.models import (
    CollectionRoles,
    CollectionTokenGrantRequest,
    CollectionTokenGrantResponse,
)


class TokenCommandTests(BaseCliTests):
    """Tests for the :py:class:`TokenCommand` class."""

    @override
    def setUp(self) -> None:
        super().setUp()
        self.grants = [
            CollectionTokenGrantResponse(
                token_type="external-user",
                token_hash=token_hash,
                username=username,
                comment="a comment",
                role=CollectionRoles.VIEWER,
            )
            for token_hash, username in (
                ("0" * 64, "test1"),
                ("1" * 64, "test2"),
            )
        ]

    @Command.preserve_registry()
    def _command(self) -> TokenCommand:
        class ConcreteTokenCommand(TokenCommand):
            """Version of TokenCommand that can be instantiated."""

            @override
            def run(self) -> None:
                raise NotImplementedError()

        return self.command(
            ConcreteTokenCommand, workspace="workspace", suite=None
        )

    def test_list_rich(self) -> None:
        command = self._command()
        with (
            mock.patch(
                "debusine.client.commands.archives.tokens.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.archives.tokens.rich.print"
            ) as rich_print,
        ):
            stderr, stdout = self.capture_output(
                partial(command._list_rich, self.grants)
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 2)
        rows = [
            call_args_list.args
            for call_args_list in table.add_row.call_args_list
        ]
        self.assertEqual(
            rows[0], ("external-user", "0" * 64, "test1", "a comment", "viewer")
        )
        self.assertEqual(
            rows[1], ("external-user", "1" * 64, "test2", "a comment", "viewer")
        )

    def test_show_rich(self) -> None:
        command = self._command()
        grant = self.grants[0].model_copy(update={"token_key": "9" * 64})
        with (
            mock.patch(
                "debusine.client.commands.archives.tokens.Table",
                return_value=(table := mock.MagicMock()),
            ),
            mock.patch(
                "debusine.client.commands.archives.tokens.rich.print"
            ) as rich_print,
        ):
            stderr, stdout = self.capture_output(
                partial(command._show_rich, grant)
            )

        self.assertEqual(stderr, "")
        self.assertEqual(stdout, "")
        rich_print.assert_called_once()

        self.assertEqual(table.add_row.call_count, 6)
        rows = [
            call_args_list.args
            for call_args_list in table.add_row.call_args_list
        ]
        self.assertEqual(rows[0], ("Token type:", "external-user"))
        self.assertEqual(rows[1], ("Token hash:", "0" * 64))
        self.assertEqual(rows[2], ("Token key (only shown once):", "9" * 64))
        self.assertEqual(rows[3], ("Username:", "test1"))
        self.assertEqual(rows[4], ("Comment:", "a comment"))
        self.assertEqual(rows[5], ("Role:", "viewer"))


class TokenGrantTests(BaseCliTests):
    """Tests for the :py:class:`TokenGrant` class."""

    def assertGrants(
        self,
        *args: str,
        collection_category: CollectionCategory,
        collection_name: str,
        request: CollectionTokenGrantRequest,
        grant: CollectionTokenGrantResponse,
    ) -> None:
        command = self.create_command(
            ["archive", "token", "grant", "--workspace=workspace", *args]
        )
        assert isinstance(command, DebusineCommand)
        with (
            mock.patch.object(
                command.debusine,
                "collection_token_grant_create",
                return_value=grant,
            ) as collection_token_grant_create,
            mock.patch.object(command, "show") as show,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        collection_token_grant_create.assert_called_once_with(
            workspace="workspace",
            collection_category=collection_category,
            collection_name=collection_name,
            request=request,
        )
        show.assert_called_once_with(grant)

    def test_archive_token_hash(self) -> None:
        self.assertGrants(
            f"--token={'0' * 64}",
            CollectionRoles.VIEWER,
            collection_category=CollectionCategory.ARCHIVE,
            collection_name="_",
            request=CollectionTokenGrantRequest(
                token_hash="0" * 64, role=CollectionRoles.VIEWER
            ),
            grant=CollectionTokenGrantResponse(
                token_type="external-user",
                token_hash="0" * 64,
                token_key="9" * 64,
                username="test",
                comment="",
                role=CollectionRoles.VIEWER,
            ),
        )

    def test_archive_username(self) -> None:
        self.assertGrants(
            "--external-username=test",
            "--comment=a comment",
            CollectionRoles.VIEWER,
            collection_category=CollectionCategory.ARCHIVE,
            collection_name="_",
            request=CollectionTokenGrantRequest(
                username="test",
                comment="a comment",
                role=CollectionRoles.VIEWER,
            ),
            grant=CollectionTokenGrantResponse(
                token_type="external-user",
                token_hash="0" * 64,
                token_key="9" * 64,
                username="test",
                comment="a comment",
                role=CollectionRoles.VIEWER,
            ),
        )

    def test_suite_token_hash(self) -> None:
        self.assertGrants(
            "--suite=sid",
            f"--token={'0' * 64}",
            CollectionRoles.VIEWER,
            collection_category=CollectionCategory.SUITE,
            collection_name="sid",
            request=CollectionTokenGrantRequest(
                token_hash="0" * 64, role=CollectionRoles.VIEWER
            ),
            grant=CollectionTokenGrantResponse(
                token_type="external-user",
                token_hash="0" * 64,
                token_key="9" * 64,
                username="test",
                comment="",
                role=CollectionRoles.VIEWER,
            ),
        )

    def test_suite_username(self) -> None:
        self.assertGrants(
            "--suite=sid",
            "--external-username=test",
            "--comment=a comment",
            CollectionRoles.VIEWER,
            collection_category=CollectionCategory.SUITE,
            collection_name="sid",
            request=CollectionTokenGrantRequest(
                username="test",
                comment="a comment",
                role=CollectionRoles.VIEWER,
            ),
            grant=CollectionTokenGrantResponse(
                token_type="external-user",
                token_hash="0" * 64,
                token_key="9" * 64,
                username="test",
                comment="a comment",
                role=CollectionRoles.VIEWER,
            ),
        )


class TokenListTests(BaseCliTests):
    """Tests for the :py:class:`TokenList` class."""

    def assertLists(
        self,
        *args: str,
        collection_category: CollectionCategory,
        collection_name: str,
    ) -> None:
        grants = [
            CollectionTokenGrantResponse(
                token_type="external-user",
                token_hash=token_hash,
                username=username,
                comment="a comment",
                role=CollectionRoles.VIEWER,
            )
            for token_hash, username in (
                ("0" * 64, "test1"),
                ("1" * 64, "test2"),
            )
        ]
        command = self.create_command(
            ["archive", "token", "list", "--workspace=workspace", *args]
        )
        assert isinstance(command, DebusineCommand)
        with (
            mock.patch.object(
                command.debusine,
                "collection_token_grant_iter",
                return_value=grants,
            ) as collection_token_grant_iter,
            mock.patch.object(command, "list") as list_,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        collection_token_grant_iter.assert_called_once_with(
            workspace="workspace",
            collection_category=collection_category,
            collection_name=collection_name,
        )
        list_.assert_called_once_with(grants)

    def test_archive(self) -> None:
        self.assertLists(
            collection_category=CollectionCategory.ARCHIVE, collection_name="_"
        )

    def test_suite(self) -> None:
        self.assertLists(
            "--suite=sid",
            collection_category=CollectionCategory.SUITE,
            collection_name="sid",
        )


class TokenRevokeTests(BaseCliTests):
    """Tests for the :py:class:`TokenRevoke` class."""

    def assertRevokes(
        self,
        *args: str,
        token_hash: str,
        role: CollectionRoles,
        collection_category: CollectionCategory,
        collection_name: str,
    ) -> None:
        command = self.create_command(
            [
                "archive",
                "token",
                "revoke",
                "--workspace=workspace",
                *args,
                token_hash,
                role,
            ]
        )
        assert isinstance(command, DebusineCommand)
        with (
            mock.patch.object(
                command.debusine, "collection_token_grant_delete"
            ) as collection_token_grant_delete,
            mock.patch.object(command, "feedback") as feedback,
        ):
            stderr, stdout = self.capture_output(command.run)

        self.assertEqual(stdout, "")
        self.assertEqual(stderr, "")
        collection_token_grant_delete.assert_called_once_with(
            workspace="workspace",
            collection_category=collection_category,
            collection_name=collection_name,
            token_hash=token_hash,
            role=role,
        )
        feedback.assert_called_once_with(
            f"Role 'viewer' revoked from {token_hash}."
        )

    def test_archive(self) -> None:
        self.assertRevokes(
            token_hash="0" * 64,
            role=CollectionRoles.VIEWER,
            collection_category=CollectionCategory.ARCHIVE,
            collection_name="_",
        )

    def test_suite(self) -> None:
        self.assertRevokes(
            "--suite=sid",
            token_hash="0" * 64,
            role=CollectionRoles.VIEWER,
            collection_category=CollectionCategory.SUITE,
            collection_name="sid",
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Common code for Debusine checks."""

import sys
from typing import Any

from django.core.checks import Error
from django.db import connections
from django.db.migrations.loader import MigrationLoader

_migration_commands = (
    "makemigrations",
    "migrate",
    "showmigrations",
    "sqlmigrate",
    "squashmigrations",
)


def is_migration_command() -> bool:
    """Return True if a migration-related command is being run."""
    return len(sys.argv) > 1 and sys.argv[1] in _migration_commands


def all_migrations_applied(admin_command: str, **kwargs: Any) -> list[Error]:
    """
    Check that all migrations have been applied.

    Avoid checking if any migration-related command is being run.

    Register this in a Django project using `django.core.checks.register`,
    passing the name of your management command as the first argument.
    """
    if is_migration_command():
        return []

    loader = MigrationLoader(connection=connections["default"], load=True)

    applied_migrations = loader.applied_migrations.keys()
    disk_migrations = loader.graph.nodes.keys()

    to_apply_migrations = sorted(disk_migrations - applied_migrations)

    return [
        Error(
            f"Non-applied migration: {migration[0]} {migration[1]}",
            hint=f'Run "{admin_command} migrate"',
        )
        for migration in to_apply_migrations
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Autopkgtest workflow."""

import datetime as dt
from collections.abc import Callable
from contextlib import suppress
from functools import cached_property
from typing import Any, TypeAlias, override

from django.utils import timezone

from debusine.artifacts import SourcePackage
from debusine.artifacts.models import (
    ArtifactCategory,
    DebianAutopkgtest,
    DebianAutopkgtestResultStatus,
)
from debusine.client.models import LookupChildType
from debusine.db.models import Artifact, WorkRequest
from debusine.server.collections.lookup import artifact_ids
from debusine.server.workflows import workflow_utils
from debusine.server.workflows.base import WorkflowValidationError
from debusine.server.workflows.models import (
    AutopkgtestWorkflowData,
    AutopkgtestWorkflowDynamicData,
    RegressionTrackingUpdateQAResults,
    WorkRequestWorkflowData,
)
from debusine.server.workflows.regression_tracking import (
    RegressionTrackingWorkflow,
)
from debusine.tasks.models import (
    ActionUpdateCollectionWithArtifacts,
    AutopkgtestData,
    AutopkgtestInput,
    BackendType,
    InputArtifact,
    InputArtifactMultiple,
    InputArtifactSingle,
    OutputData,
    RegressionAnalysis,
    RegressionAnalysisStatus,
)


class AutopkgtestWorkflow(
    RegressionTrackingWorkflow[
        AutopkgtestWorkflowData, AutopkgtestWorkflowDynamicData
    ]
):
    """Run autopkgtests for a single source on a set of architectures."""

    TASK_NAME = "autopkgtest"

    @override
    def post_init(self) -> None:
        super().post_init()
        if self.data.backend == BackendType.AUTO:
            self.data.backend = BackendType.UNSHARE

    @cached_property
    def architectures(self) -> set[str]:
        """Concrete architectures to run tests on."""
        try:
            architectures = workflow_utils.get_lookup_architectures(
                self, self.data.binary_artifacts
            )
        except ValueError as e:
            raise WorkflowValidationError(str(e)) from e

        # If the input only contains architecture-independent binary
        # packages, then run tests on {arch_all_build_architecture}.  If it
        # contains a mix of architecture-independent and
        # architecture-dependent binary packages, then running tests on any
        # of the concrete architectures will do.
        if architectures == {"all"}:
            architectures = {self.data.arch_all_build_architecture}
        architectures.discard("all")

        if self.data.architectures:
            architectures &= set(self.data.architectures)

        return architectures

    @override
    def validate_input(self) -> None:
        """Thorough validation of input data."""
        super().validate_input()
        # binary_artifacts is validated by accessing self.architectures.
        self.architectures

    def _has_current_reference_qa_result(self, architecture: str) -> bool:
        """
        Return True iff we have a current reference QA result.

        An autopkgtest analysis is outdated if:

        * either the underlying source or binary packages are outdated (i.e.
          have different version numbers) compared to what's available in
          the ``debian:suite`` collection
        * or the timestamp of the analysis is older than 30 days compared to
          the ``Date`` timestamp of the ``debian:suite`` collection

        Otherwise, it is current.
        """
        # This method is only called when update_qa_results is "yes", in
        # which case this is checked by a model validator.
        assert self.reference_qa_results is not None

        source_data = workflow_utils.source_package_data(self)
        latest_result = self.reference_qa_results.manager.lookup(
            f"latest:autopkgtest_{source_data.name}_{architecture}"
        )
        return (
            latest_result is not None
            and latest_result.data["version"] == source_data.version
            and (
                self.qa_suite_changed is None
                or dt.datetime.fromtimestamp(
                    latest_result.data["timestamp"], tz=dt.UTC
                )
                >= self.qa_suite_changed - dt.timedelta(days=30)
            )
        )

    def _populate_single(self, architecture: str) -> None:
        """Create an autopkgtest work request for a single architecture."""
        if (
            self.data.update_qa_results == RegressionTrackingUpdateQAResults.YES
            and self._has_current_reference_qa_result(architecture)
        ):
            return

        source_artifact = workflow_utils.locate_debian_source_package_lookup(
            self, "source_artifact", self.data.source_artifact
        )
        filtered_binary_artifacts = (
            workflow_utils.filter_artifact_lookup_by_arch(
                self, self.data.binary_artifacts, {architecture, "all"}
            )
        )
        filtered_context_artifacts = (
            workflow_utils.filter_artifact_lookup_by_arch(
                self, self.data.context_artifacts, {architecture, "all"}
            )
        )

        environment = f"{self.data.vendor}/match:codename={self.data.codename}"
        extra_repositories = workflow_utils.configure_for_overlay_suite(
            self,
            extra_repositories=self.data.extra_repositories,
            vendor=self.data.vendor,
            codename=self.data.codename,
            environment=environment,
            backend=self.data.backend,
            architecture=architecture,
            try_variant="autopkgtest",
        )

        task_data = AutopkgtestData(
            input=AutopkgtestInput(
                source_artifact=source_artifact,
                binary_artifacts=filtered_binary_artifacts,
                context_artifacts=filtered_context_artifacts,
            ),
            build_architecture=architecture,
            environment=environment,
            backend=self.data.backend,
            include_tests=self.data.include_tests,
            exclude_tests=self.data.exclude_tests,
            extra_repositories=extra_repositories,
            debug_level=self.data.debug_level,
            pin_packages=self.data.pin_packages,
            extra_environment=self.data.extra_environment,
            needs_internet=self.data.needs_internet,
            fail_on=self.data.fail_on,
            timeout=self.data.timeout,
        )
        workflow_data_step = f"autopkgtest-{architecture}"
        workflow_data_kwargs: dict[str, Any] = {}
        if self.data.update_qa_results != RegressionTrackingUpdateQAResults.NO:
            # When updating reference results for regression tracking, task
            # failures never cause the parent workflow or dependent tasks to
            # fail.
            workflow_data_kwargs["allow_failure"] = True
        wr, created = self.work_request_ensure_child_worker(
            task_name="autopkgtest",
            task_data=task_data,
            workflow_data=WorkRequestWorkflowData(
                display_name=f"autopkgtest {architecture}",
                step=workflow_data_step,
                **workflow_data_kwargs,
            ),
        )
        self.requires_artifact(wr, source_artifact)
        self.requires_artifact(wr, filtered_binary_artifacts)
        self.requires_artifact(wr, filtered_context_artifacts)
        promise_name = f"{self.data.prefix}autopkgtest-{architecture}"
        self.provides_artifact(wr, ArtifactCategory.AUTOPKGTEST, promise_name)

        if created and (
            self.data.update_qa_results != RegressionTrackingUpdateQAResults.NO
        ):
            # Checked by a model validator.
            assert self.data.reference_qa_results is not None

            source_data = workflow_utils.source_package_data(self)

            # Back off if another workflow gets there first.
            self.skip_if_qa_result_changed(
                wr,
                package=source_data.name,
                architecture=architecture,
                promise_name=promise_name,
            )

            # Record results in the reference collection.
            action = ActionUpdateCollectionWithArtifacts(
                collection=self.data.reference_qa_results,
                variables={
                    "package": source_data.name,
                    "version": source_data.version,
                    "architecture": architecture,
                    "timestamp": int(
                        (self.qa_suite_changed or timezone.now()).timestamp()
                    ),
                    "work_request_id": wr.id,
                },
                artifact_filters={"category": ArtifactCategory.AUTOPKGTEST},
            )
            wr.add_event_reaction("on_success", action)
            wr.add_event_reaction("on_failure", action)

    def _populate_regression_analysis_callback(
        self, architectures: set[str]
    ) -> None:
        """Create a callback to collate regression analysis results."""
        if self.data.enable_regression_tracking:
            # Checked by a model validator.
            assert self.data.reference_prefix

            regression_analysis, _ = self.work_request_ensure_child_internal(
                task_name="workflow",
                workflow_data=WorkRequestWorkflowData(
                    allow_dependency_failures=True,
                    step="regression-analysis",
                    display_name="Regression analysis",
                    visible=False,
                ),
                unblock_strategy=WorkRequest.UnblockStrategy.INCREMENTAL,
            )
            for architecture in sorted(architectures):
                for prefix in (self.data.reference_prefix, self.data.prefix):
                    with suppress(KeyError):
                        self.requires_artifact(
                            regression_analysis,
                            f"internal@collections/name:"
                            f"{prefix}autopkgtest-{architecture}",
                        )
            self.orchestrate_child(regression_analysis)

    @override
    def populate(self) -> None:
        """Create autopkgtest work requests for all architectures."""
        for architecture in sorted(self.architectures):
            self._populate_single(architecture)
        self._populate_regression_analysis_callback(self.architectures)

    def _extract_artifact_details(
        self, artifact: Artifact | None
    ) -> tuple[str | None, int | None, WorkRequest | None]:
        """Extract details from a ``debian:autopkgtest`` artifact."""
        source_version: str | None
        artifact_id: int | None
        wr: WorkRequest | None
        if artifact is None:
            source_version = None
            artifact_id = None
            wr = None
        else:
            data = artifact.create_data()
            assert isinstance(data, DebianAutopkgtest)
            source_version = data.source_package.version
            artifact_id = artifact.id
            wr = artifact.created_by_work_request
        return source_version, artifact_id, wr

    @staticmethod
    def compare_qa_results_fine_grained(
        reference: Artifact | None, new: Artifact | None
    ) -> dict[str, str]:
        """Do a fine-grained comparison of two QA results."""
        reference_results = (
            {} if reference is None else reference.data["results"]
        )
        new_results = {} if new is None else new.data["results"]
        DARS: TypeAlias = DebianAutopkgtestResultStatus
        details = {}
        for test_name in set(reference_results) | set(new_results):
            match reference_results.get(test_name), new_results.get(test_name):
                case (
                    {"status": DARS.PASS | DARS.SKIP} | None,
                    {"status": DARS.FAIL},
                ):
                    details[test_name] = "regression"
                case (
                    {"status": DARS.FAIL | DARS.FLAKY},
                    {"status": DARS.PASS | DARS.SKIP} | None,
                ):
                    details[test_name] = "improvement"
                case _:
                    details[test_name] = "stable"
        return details

    def callback_regression_analysis(self) -> bool:
        """
        Analyze regressions compared to reference results.

        This is called once for each architecture, but updates the whole
        analysis for all architectures each time.  This is partly for
        simplicity and robustness (we don't need to work out how to combine
        the new analysis with a previous one), and partly to make it easier
        to handle cases where there isn't a one-to-one mapping between the
        reference results and the new results.
        """
        # Select the newest result for each architecture.
        reference_artifacts = self.find_reference_artifacts("autopkgtest")
        new_artifacts = self.find_new_artifacts(
            "autopkgtest", ArtifactCategory.AUTOPKGTEST
        )

        output_data = self.work_request.output_data or OutputData()
        output_data.regression_analysis = {}
        for architecture in sorted(
            set(reference_artifacts) | set(new_artifacts)
        ):
            reference = reference_artifacts.get(architecture)
            reference_source_version, reference_artifact_id, reference_wr = (
                self._extract_artifact_details(reference)
            )
            new = new_artifacts.get(architecture)
            new_source_version, new_artifact_id, new_wr = (
                self._extract_artifact_details(new)
            )

            status = self.compare_qa_results(reference_wr, new_wr)
            details = self.compare_qa_results_fine_grained(reference, new)
            if status == RegressionAnalysisStatus.STABLE:
                if "regression" in details.values():
                    status = RegressionAnalysisStatus.REGRESSION
                elif "improvement" in details.values():
                    status = RegressionAnalysisStatus.IMPROVEMENT
                else:
                    status = RegressionAnalysisStatus.STABLE

            output_data.regression_analysis[architecture] = RegressionAnalysis(
                original_source_version=reference_source_version,
                original_artifact_id=reference_artifact_id,
                new_source_version=new_source_version,
                new_artifact_id=new_artifact_id,
                status=status,
                details=details,
            )

        self.work_request.output_data = output_data
        self.work_request.save()
        return True

    @override
    def get_callback(self, step: str) -> Callable[[], bool]:
        # Current versions of this workflow create a single incremental
        # callback for all architectures called "regression-analysis", but
        # we may be run on workflows that were orchestrated when we created
        # one "regression-analysis-{architecture}" callback per
        # architecture.
        if not step.startswith("regression-analysis"):
            raise NotImplementedError(
                f"Unhandled workflow callback step: {step}"
            )
        return self.callback_regression_analysis

    @override
    def compute_dynamic_data(self) -> AutopkgtestWorkflowDynamicData:
        """
        Compute dynamic data for this workflow.

        :subject: package name of ``source_artifact``
        :parameter_summary: package name_version
        :source_artifact_id: id of the source artifact
        :binary_artifacts_ids: ids of the binary artifacts
        :context_artifacts_ids: ids of the context artifacts
        """
        source_artifact = workflow_utils.source_package(self)
        source_data = SourcePackage.create_data(source_artifact.data)

        binary_artifacts_results = self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )
        context_artifacts_results = self.work_request.lookup_multiple(
            self.data.context_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )

        return AutopkgtestWorkflowDynamicData(
            subject=source_data.name,
            parameter_summary=f"{source_data.name}_{source_data.version}",
            source_artifact_id=source_artifact.id,
            binary_artifacts_ids=artifact_ids(binary_artifacts_results),
            context_artifacts_ids=artifact_ids(context_artifacts_results),
        )

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return input artifacts."""
        return [
            InputArtifactSingle(
                lookup=self.data.source_artifact,
                label="source_artifact",
                artifact_id=(
                    self.dynamic_data.source_artifact_id
                    if self.dynamic_data is not None
                    else None
                ),
            ),
            InputArtifactMultiple.from_ids(
                lookup=self.data.binary_artifacts,
                label="binary_artifacts",
                artifact_ids=(
                    self.dynamic_data.binary_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            ),
            InputArtifactMultiple.from_ids(
                lookup=self.data.context_artifacts,
                label="context_artifacts",
                artifact_ids=(
                    self.dynamic_data.context_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            ),
        ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.
"""Debusine extensions to Signon."""

import logging
from typing import Any, override

import requests
from rest_framework import status

from debusine.db.models import Identity, User
from debusine.server.signon import providers
from debusine.server.signon.signon import Signon
from debusine.utils import get_user_agent
from debusine.utils.urls import DEFAULT_REQUEST_TIMEOUT

log = logging.getLogger("debusine.server.signon")


class DebianSignon(Signon):
    """
    Debian-specific extension to Signon.

    Activate it in django settings with::

        SIGNON_CLASS = "debusine.server.signon.sites.DebianSignon"

    This replaces the standard user authorization with a custom one for Debian,
    which:

    * ``restrict`` is ignored.
    * Fetches user status from nm.debian.org add ads it to claims as
      ``nm:status``
    * Checks for the ``debian`` group in Salsa claims: if present,
      authorization succeeds
    * Check for ``dm`` or ``dm_ga`` statuses in ``nm:status``: if present,
      authorization succeeds
    * ``add_to_group`` is extended to match ``nm:status`` values as groups,
      when prefixed with ``nm:``: see example below.

    You can then set a provider to auto-add users to Debusine groups with::

        SIGNON_PROVIDERS = [
            providers.ProviderClass(
                ...,
                options={
                    "nm_timeout": (5, 30),  # (connect, read)
                    "add_to_group": {
                        # Map gitlab group names to Debusine groups
                        "gitlab/group": "scopename/groupname",
                        # Map nm.debian.org statuses to Debusine groups
                        "nm:dm": "scopename/groupname",
                        "nm:dm_ga": "scopename/groupname",
                    },
                }
            )
        ]
    """

    def fetch_nm_claims(
        self,
        salsa_id: int,
        timeout: int | tuple[int, int] = DEFAULT_REQUEST_TIMEOUT,
    ) -> dict[str, Any]:
        """Fetch user information from nm.debian.org."""
        url = f"https://nm.debian.org/api/salsa_status/{salsa_id}"

        with requests.get(
            url, headers={"User-Agent": get_user_agent()}, timeout=timeout
        ) as r:
            match r.status_code:
                case status.HTTP_200_OK:
                    if nm_status := r.json().get("status"):
                        return {"nm_status": nm_status}
                    log.warning(
                        "%s: user record has no 'status' information", url
                    )
                    return {}
                case status.HTTP_404_NOT_FOUND:
                    log.info("%s: user is not known in nm.debian.org", url)
                    return {}
                case _:
                    log.warning(
                        "%s: cannot fetch user information: %d (%s)",
                        url,
                        r.status_code,
                        r.reason,
                    )
                    return {}

    @override
    def validate_claims(
        self, provider: providers.Provider, identity: "Identity"
    ) -> list[str]:
        """
        Check if the claims in identity are acceptable.

        :return: a list of error messages, or an empty list if validation passed
        """
        res = []
        if not isinstance(provider, providers.GitlabProvider):
            res.append(
                "unsupported validation of logins from"
                f" {provider.name!r} provider:"
                f" {provider.__class__.__name__} is not supported"
            )
            return res

        # Get timeout from admin options, fallback to global constant
        nm_timeout = provider.options.get("nm_timeout", DEFAULT_REQUEST_TIMEOUT)

        # Work only with verified emails
        if not identity.claims.get("email_verified", False):
            res.append(f"identity {identity} does not have a verified email")
            return res

        # Call out to nm.debian.org to add claims["nm_status"]
        if nm_claims := self.fetch_nm_claims(
            int(identity.subject), timeout=nm_timeout
        ):
            identity.claims.update(nm_claims)
            identity.save()

        if "debian" in identity.claims.get("groups_direct", ()):
            # Allow if in the Debian group.
            # We could check nm_status here, but checking groups_direct allows
            # creating accounts for DDs even if nm.debian.org is having
            # troubles
            return res
        elif identity.claims.get("nm_status") in ("dm", "dm_ga"):
            # Allow if a DM / DM with guest account
            return res
        else:
            res.append(f"identity {identity} is not DD or DM")
            return res

    @override
    def add_user_to_groups(
        self, user: User, identity: Identity, add_to_group: dict[str, str]
    ) -> None:
        """Add a user to groups according to ``add_to_group`` rules."""
        gitlab_groups = identity.claims.get("groups_direct", ())
        nm_status = identity.claims.get("nm_status", "")
        for remote_group, debusine_group in add_to_group.items():
            if remote_group.startswith("nm:"):
                if nm_status != remote_group[3:]:
                    continue
            elif remote_group not in gitlab_groups:
                continue
            self.add_user_to_group(user, debusine_group)



# Generated by Django 3.2.19 on 2024-07-12 13:26

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('signing', '0002_key_discriminated_union'),
    ]

    operations = [
        migrations.AlterField(
            model_name='auditlog',
            name='created_by_work_request_id',
            field=models.IntegerField(blank=True, null=True),
        ),
        migrations.AlterField(
            model_name='auditlog',
            name='event',
            field=models.CharField(
                choices=[
                    ('generate', 'Generate'),
                    ('sign', 'Sign'),
                    ('register', 'Register an externally-generated key'),
                ],
                max_length=8,
            ),
        ),
    ]



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests that try exploiting the system via its public API."""

from typing import Any, ClassVar, override

from django.urls import reverse
from rest_framework import status
from rest_framework.test import APIClient

from debusine.assets import KeyPurpose as ArtifactKeyPurpose
from debusine.assets import SigningKeyData
from debusine.db.models import (
    Asset,
    AssetUsage,
    Token,
    WorkRequest,
    WorkflowTemplate,
    Workspace,
)
from debusine.db.playground import scenarios
from debusine.server.celery import run_server_task
from debusine.server.workflows.base import orchestrate_workflow
from debusine.signing.tasks.models import SignData, SignDynamicData
from debusine.tasks.models import LookupMultiple, WorkerType
from debusine.test.django import TestCase, TestResponseType


class PenetrationTests(TestCase):
    """Try to game the system using its API."""

    scenario = scenarios.DefaultContextAPI()

    create_child_workspace: ClassVar[WorkflowTemplate]
    signing_key: ClassVar[Asset]
    signing_key_usage: ClassVar[AssetUsage]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        # Allow creation of child workspaces
        cls.create_child_workspace = cls.playground.create_workflow_template(
            "create_child", "create_child_workspace"
        )
        cls.signing_key = cls.playground.create_signing_key_asset()
        cls.signing_key_usage = cls.playground.create_asset_usage(
            cls.signing_key
        )

    @override
    def setUp(self) -> None:
        super().setUp()
        self.client = APIClient()

    def _post(
        self,
        endpoint: str,
        *,
        data: dict[str, Any],
        endpoint_kwargs: dict[str, Any] | None = None,
        scope: str,
        token: Token,
    ) -> TestResponseType:
        headers = {"Token": token.key}
        headers["X-Debusine-Scope"] = scope

        return self.client.post(
            reverse(endpoint, kwargs=endpoint_kwargs),
            data=data,
            headers=headers,
            format="json",
        )

    def post_workflow(self, data: dict[str, Any]) -> TestResponseType:
        """Post a workflow creation request to api:workflows."""
        return self._post(
            "api:workflows",
            data=data,
            scope=self.scenario.scope.name,
            token=self.scenario.user_token,
        )

    def post_permission_check(
        self,
        *,
        asset_category: str,
        asset_slug: str,
        permission_name: str,
        data: dict[str, Any],
        work_request: WorkRequest,
    ) -> TestResponseType:
        """Post an asset permission check request."""
        work_request.worker = self.playground.create_worker(
            worker_type=WorkerType.SIGNING
        )
        work_request.save()
        worker_token = self.playground.create_work_request_token(work_request)
        return self._post(
            "api:asset-permission-check",
            endpoint_kwargs={
                "asset_category": asset_category,
                "asset_slug": asset_slug,
                "permission_name": permission_name,
            },
            data=data,
            scope=self.scenario.scope.name,
            token=worker_token,
        )

    def assertChildWorkspaceCreated(self, **kwargs: Any) -> Workspace:
        """Create a child workspace and return its name."""
        response = self.post_workflow(
            {
                "template_name": "create_child",
                "task_data": kwargs,
            }
        )
        result = self.assertAPIResponseOk(response, status.HTTP_201_CREATED)

        # Run the workflow
        workflow = WorkRequest.objects.get(pk=result["id"])
        orchestrate_workflow(workflow)

        # Run the server task
        server_task = workflow.children.get()
        self.playground.advance_work_request(
            server_task, assign_new_worker=True
        )
        run_server_task.apply(args=(server_task.pk,))
        server_task.refresh_from_db()
        self.assertEqual(server_task.status, WorkRequest.Statuses.COMPLETED)
        self.assertEqual(server_task.result, WorkRequest.Results.SUCCESS)

        return Workspace.objects.get(
            scope=self.scenario.scope,
            name=f"{self.scenario.workspace.name}-test",
        )

    def test_use_signing_key_from_child(self) -> None:
        """Attempt to use a workflow signing key in a child workspace."""
        self.playground.create_group_role(
            self.scenario.workspace,
            Workspace.Roles.CONTRIBUTOR,
            users=[self.scenario.user],
        )
        new_workspace = self.assertChildWorkspaceCreated(suffix="test")
        signing_input = self.playground.create_signing_input_artifact(
            workspace=new_workspace
        )
        signing_key_data = self.signing_key.data_model
        assert isinstance(signing_key_data, SigningKeyData)

        signing_task = self.playground.create_signing_task(
            workspace=new_workspace,
            task_name="sign",
            task_data=SignData(
                purpose=ArtifactKeyPurpose(signing_key_data.purpose),
                unsigned=LookupMultiple((signing_input.id,)),
                key=signing_key_data.fingerprint,
            ),
            created_by=self.scenario.user,
            mark_running=True,
        )
        signing_task.dynamic_task_data = SignDynamicData(
            unsigned_ids=[signing_input.id],
            unsigned_binary_package_names=[
                signing_input.data["binary_package_name"]
            ],
        ).model_dump()
        signing_task.save()
        response = self.post_permission_check(
            asset_category=self.signing_key.category,
            asset_slug=self.signing_key.slug,
            permission_name="sign_with",
            data={
                "artifact_id": signing_input.id,
                "work_request_id": signing_task.id,
                "workspace": new_workspace.name,
            },
            work_request=signing_task,
        )
        result = self.assertAPIResponseOk(response)
        self.assertFalse(result["has_permission"])



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Tests the artifacts UI helper."""

from typing import ClassVar, override

from django.test import RequestFactory

from debusine.db.models import Artifact, FileInArtifact
from debusine.db.playground import scenarios
from debusine.test.django import TestCase
from debusine.web.views.ui.artifacts import ArtifactUI, FileInArtifactUI


class ArtifactUITests(TestCase):
    """Test the :py:class:`ArtifactUI` class."""

    scenario = scenarios.DefaultContext(set_current=True)

    artifact: ClassVar[Artifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        cls.artifact, _ = cls.playground.create_artifact()

    def ui(self) -> ArtifactUI:
        """Instantiate the UI helper to test."""
        factory = RequestFactory()
        request = factory.get("/")
        return self.artifact.ui(request)

    def test_place(self) -> None:
        place = self.ui().place
        self.assertEqual(place.title, f"Artifact {self.artifact.get_label()}")
        self.assertEqual(place.url, self.artifact.get_absolute_url())
        self.assertEqual(place.icon, "file")
        self.assertEqual(
            place.description, "debusine:test artifact debusine:test"
        )
        self.assertEqual(place.breadcrumb, "debusine:test")
        assert place.parent is not None
        self.assertEqual(
            place.parent.url, self.scenario.workspace.get_absolute_url()
        )


class FileInArtifactUITests(TestCase):
    """Test the :py:class:`FileInArtifactUI` class."""

    scenario = scenarios.DefaultContext(set_current=True)

    file: ClassVar[FileInArtifact]

    @override
    @classmethod
    def setUpTestData(cls) -> None:
        super().setUpTestData()
        artifact, _ = cls.playground.create_artifact(
            {"hello.deb": b"contents"}, create_files=True
        )
        cls.file = FileInArtifact.objects.get(
            artifact=artifact, path="hello.deb"
        )

    def ui(self) -> FileInArtifactUI:
        """Instantiate the UI helper to test."""
        factory = RequestFactory()
        request = factory.get("/")
        return self.file.ui(request)

    def test_place(self) -> None:
        place = self.ui().place
        self.assertEqual(place.title, self.file.path)
        self.assertEqual(place.url, self.file.get_absolute_url())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)
        self.assertEqual(place.breadcrumb, "hello.deb")
        assert place.parent is not None
        self.assertEqual(
            place.parent.url, self.file.artifact.get_absolute_url()
        )

    def test_place_incomplete(self) -> None:
        self.file.complete = False
        place = self.ui().place
        self.assertEqual(place.title, self.file.path + " (incomplete)")
        self.assertEqual(place.url, self.file.get_absolute_url())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)
        self.assertEqual(place.breadcrumb, "hello.deb")
        assert place.parent is not None
        self.assertEqual(
            place.parent.url, self.file.artifact.get_absolute_url()
        )

    def test_place_raw(self) -> None:
        place = self.ui().place_raw
        self.assertEqual(place.title, self.file.path)
        self.assertEqual(place.url, self.file.get_absolute_url_raw())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)
        self.assertEqual(place.breadcrumb, "hello.deb")
        assert place.parent is not None
        self.assertEqual(
            place.parent.url, self.file.artifact.get_absolute_url()
        )

    def test_place_raw_incomplete(self) -> None:
        self.file.complete = False
        place = self.ui().place_raw
        self.assertEqual(place.title, self.file.path + " (incomplete)")
        self.assertEqual(place.url, self.file.get_absolute_url_raw())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)
        self.assertEqual(place.breadcrumb, "hello.deb")
        assert place.parent is not None
        self.assertEqual(
            place.parent.url, self.file.artifact.get_absolute_url()
        )

    def test_place_download(self) -> None:
        place = self.ui().place_download
        self.assertEqual(place.title, self.file.path)
        self.assertEqual(place.url, self.file.get_absolute_url_download())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)
        self.assertEqual(place.breadcrumb, "hello.deb")
        assert place.parent is not None
        self.assertEqual(
            place.parent.url, self.file.artifact.get_absolute_url()
        )

    def test_place_download_incomplete(self) -> None:
        self.file.complete = False
        place = self.ui().place_download
        self.assertEqual(place.title, self.file.path + " (incomplete)")
        self.assertEqual(place.url, self.file.get_absolute_url_download())
        self.assertIsNone(place.icon)
        self.assertIsNone(place.description)
        self.assertEqual(place.breadcrumb, "hello.deb")
        assert place.parent is not None
        self.assertEqual(
            place.parent.url, self.file.artifact.get_absolute_url()
        )



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Package publishing workflow."""

import logging
from typing import Any, override

from debusine.artifacts.models import (
    ArtifactCategory,
    CollectionCategory,
    DebianBinaryPackage,
    DebianSourcePackage,
    DebianUpload,
)
from debusine.client.models import LookupChildType, RuntimeParameter
from debusine.db.models import (
    Artifact,
    Collection,
    WellKnownWorkflowTemplate,
    WorkRequest,
    WorkflowTemplate,
    Workspace,
)
from debusine.db.tasks.inputs import CollectionInSameWorkspace
from debusine.server.collections.lookup import artifact_ids
from debusine.server.tasks.models import (
    CopyCollectionItemsCopies,
    CopyCollectionItemsData,
)
from debusine.server.workflows import Workflow, workflow_utils
from debusine.server.workflows.models import (
    PackagePublishWorkflowData,
    PackagePublishWorkflowDynamicData,
    UpdateSuitesData,
    WorkRequestWorkflowData,
)
from debusine.tasks import TaskConfigError
from debusine.tasks.inputs import Stage
from debusine.tasks.models import (
    InputArtifact,
    InputArtifactMultiple,
    InputArtifactSingle,
    LookupMultiple,
)

logger = logging.getLogger(__name__)


class PackagePublishWorkflow(
    Workflow[PackagePublishWorkflowData, PackagePublishWorkflowDynamicData]
):
    """Publish packages to a target suite."""

    TASK_NAME = "package_publish"

    target_suite = CollectionInSameWorkspace(
        category=CollectionCategory.SUITE, stage=Stage.PENDING
    )

    def _split_component_and_section(
        self, component_and_section: str
    ) -> tuple[str, str]:
        """
        Split a ``Section`` field into component and section.

        See
        https://www.debian.org/doc/debian-policy/ch-archive.html#s-subsections
        (which refers to components as "archive areas").
        """
        if "/" in component_and_section:
            component, section = component_and_section.split("/", 1)
            return component, section
        else:
            return "main", component_and_section

    def _add_source(
        self, copies: list[CopyCollectionItemsCopies], target_suite: Collection
    ) -> None:
        """Add a copy item for the source artifact."""
        if self.data.source_artifact is None:
            return

        original_source_artifact = self.work_request.lookup_single(
            self.data.source_artifact, expect_type=LookupChildType.ARTIFACT
        ).artifact
        source_artifact = workflow_utils.locate_debian_source_package(
            "source_artifact", original_source_artifact
        )
        variables: dict[str, Any] = {}

        # Set the component and section.
        # TODO: We should use the following logic:
        #  * use the current overrides in the target suite, if any
        #  * if the source artifact was looked up via a collection, use the
        #    current overrides from there, if any
        #  * if the source artifact is an upload, use the component and
        #    section listed for its `.dsc` in its `.changes` file
        #  * otherwise, default to `main`/`misc`
        if original_source_artifact.category == ArtifactCategory.UPLOAD:
            # Extract reasonable component and section defaults from the
            # metadata for the .dsc file in the .changes file.
            original_source_data = original_source_artifact.create_data()
            assert isinstance(original_source_data, DebianUpload)
            [dsc_section] = [
                file["section"]
                for file in original_source_data.changes_fields["Files"]
                if file["name"].endswith(".dsc")
            ]
            variables["component"], variables["section"] = (
                self._split_component_and_section(dsc_section)
            )
        else:
            # If we don't have a .changes file, then we can't easily get at
            # the source package's component or section defaults (at least
            # not without unpacking the source package), so the best we can
            # do is to pick some generic defaults.
            source_data = source_artifact.create_data()
            assert isinstance(source_data, DebianSourcePackage)
            # TODO: These log messages aren't visible to users; see
            # https://salsa.debian.org/freexian-team/debusine/-/issues/901.
            if "component" not in self.data.suite_variables:
                logger.warning(
                    "No component specified when adding %s to %s; "
                    "defaulting to 'main'",
                    source_data.get_label(),
                    self.data.target_suite,
                )
                variables["component"] = "main"
            if "section" not in self.data.suite_variables:
                logger.warning(
                    "No section specified when adding %s to %s; "
                    "defaulting to 'misc'",
                    source_data.get_label(),
                    self.data.target_suite,
                )
                variables["section"] = "misc"

        copies.append(
            CopyCollectionItemsCopies(
                source_items=LookupMultiple(
                    (f"{source_artifact.id}@artifacts",)
                ),
                # Identify the target collection by ID to avoid strange
                # behaviour if the lookup result changes between now and
                # when the child task executes, e.g. due to creating a
                # collection earlier in the inheritance chain.  Implementing
                # https://salsa.debian.org/freexian-team/debusine/-/issues/495
                # may allow us to do better here.
                target_collection=target_suite.id,
                unembargo=self.data.unembargo,
                replace=self.data.replace,
                variables=variables | self.data.suite_variables,
            )
        )

    def _add_binary(
        self,
        copies: list[CopyCollectionItemsCopies],
        target_suite: Collection,
        binary_artifact: Artifact,
    ) -> None:
        """Add copy items for binary artifacts."""
        binary_data = binary_artifact.create_data()
        assert isinstance(binary_data, DebianBinaryPackage)
        variables: dict[str, Any] = {}

        # TODO: We should use the following logic:
        #  * use the current overrides in the target suite, if any
        #  * if a binary artifact was looked up via a collection, use the
        #    current overrides from there, if any
        #  * if a binary artifact has `Section`/`Priority` fields in its
        #    `.deb`, use those
        #  * otherwise, default to `main`/`misc`/`optional`
        if "Section" in binary_data.deb_fields:
            variables["component"], variables["section"] = (
                self._split_component_and_section(
                    binary_data.deb_fields["Section"]
                )
            )
        else:
            # TODO: These log messages aren't visible to users; see
            # https://salsa.debian.org/freexian-team/debusine/-/issues/901.
            if "component" not in self.data.suite_variables:
                logger.warning(
                    "No component specified when adding %s to %s; "
                    "defaulting to 'main'",
                    binary_data.get_label(),
                    self.data.target_suite,
                )
                variables["component"] = "main"
            if "section" not in self.data.suite_variables:
                logger.warning(
                    "No section specified when adding %s to %s; "
                    "defaulting to 'misc'",
                    binary_data.get_label(),
                    self.data.target_suite,
                )
                variables["section"] = "misc"

        if "Priority" in binary_data.deb_fields:
            variables["priority"] = binary_data.deb_fields["Priority"]
        elif "priority" not in self.data.suite_variables:
            # TODO: This log message isn't visible to users; see
            # https://salsa.debian.org/freexian-team/debusine/-/issues/901.
            logger.warning(
                "No priority specified when adding %s to %s; "
                "defaulting to 'optional'",
                binary_data.get_label(),
                self.data.target_suite,
            )
            variables["priority"] = "optional"

        copies.append(
            CopyCollectionItemsCopies(
                source_items=LookupMultiple(
                    (f"{binary_artifact.id}@artifacts",)
                ),
                # Identify the target collection by ID to avoid strange
                # behaviour if the lookup result changes between now and
                # when the child task executes, e.g. due to creating a
                # collection earlier in the inheritance chain.  Implementing
                # https://salsa.debian.org/freexian-team/debusine/-/issues/495
                # may allow us to do better here.
                target_collection=target_suite.id,
                unembargo=self.data.unembargo,
                replace=self.data.replace,
                variables=variables | self.data.suite_variables,
            )
        )

    def _add_build_logs(
        self,
        copies: list[CopyCollectionItemsCopies],
        *,
        source_workspace: Workspace,
        target_workspace: Workspace,
    ) -> None:
        """Add copy items for build logs."""
        try:
            source_build_logs_collection = self.lookup_singleton_collection(
                CollectionCategory.PACKAGE_BUILD_LOGS,
                workspace=source_workspace,
            )
        except KeyError:
            source_build_logs_collection = None
        try:
            target_build_logs_collection = self.lookup_singleton_collection(
                CollectionCategory.PACKAGE_BUILD_LOGS,
                workspace=target_workspace,
            )
        except KeyError:
            target_build_logs_collection = None
        if (
            source_build_logs_collection is not None
            and target_build_logs_collection is not None
            and source_build_logs_collection != target_build_logs_collection
        ):
            copies.append(
                CopyCollectionItemsCopies(
                    source_items=LookupMultiple.model_validate(
                        {
                            "collection": source_build_logs_collection.id,
                            "lookup__same_work_request": (
                                self.data.binary_artifacts
                            ),
                        }
                    ),
                    target_collection=target_build_logs_collection.id,
                    unembargo=self.data.unembargo,
                    replace=self.data.replace,
                )
            )

    def _add_task_history(
        self,
        copies: list[CopyCollectionItemsCopies],
        *,
        source_workspace: Workspace,
        target_workspace: Workspace,
    ) -> None:
        """Add copy items for task history."""
        try:
            source_task_history_collection = self.lookup_singleton_collection(
                CollectionCategory.TASK_HISTORY, workspace=source_workspace
            )
        except KeyError:
            source_task_history_collection = None
        try:
            target_task_history_collection = self.lookup_singleton_collection(
                CollectionCategory.TASK_HISTORY, workspace=target_workspace
            )
        except KeyError:
            target_task_history_collection = None
        if (
            source_task_history_collection is not None
            and target_task_history_collection is not None
            and source_task_history_collection != target_task_history_collection
        ):
            copies.append(
                CopyCollectionItemsCopies(
                    source_items=LookupMultiple.model_validate(
                        {
                            "collection": source_task_history_collection.id,
                            "lookup__same_workflow": self.data.binary_artifacts,
                        }
                    ),
                    target_collection=target_task_history_collection.id,
                    unembargo=self.data.unembargo,
                    replace=self.data.replace,
                )
            )

    @override
    def populate(self) -> None:
        """Create work requests."""
        assert self.dynamic_data is not None
        subject = self.dynamic_data.subject
        assert subject is not None

        target_suite = self.target_suite.collection
        target_workspace = target_suite.workspace

        copies: list[CopyCollectionItemsCopies] = []
        self._add_source(copies, target_suite)
        binary_artifacts_workspace: Workspace | None = None
        for binary_artifact in workflow_utils.locate_debian_binary_packages(
            "binary_artifacts",
            [
                result.artifact
                for result in self.work_request.lookup_multiple(
                    self.data.binary_artifacts,
                    expect_type=LookupChildType.ARTIFACT,
                )
            ],
        ):
            self._add_binary(copies, target_suite, binary_artifact)
            if binary_artifacts_workspace is None:
                binary_artifacts_workspace = binary_artifact.workspace
            elif binary_artifact.workspace != binary_artifacts_workspace:
                # TODO: This should be handled during the configuration
                # stage, but we first need to have input fields that can
                # cope with promises.
                raise TaskConfigError(
                    f"Binary artifacts must all be in the same workspace"
                    f" ({binary_artifact.workspace} !="
                    f" {binary_artifacts_workspace})"
                )
        if binary_artifacts_workspace is not None:
            self._add_build_logs(
                copies,
                source_workspace=binary_artifacts_workspace,
                target_workspace=target_workspace,
            )
            self._add_task_history(
                copies,
                source_workspace=binary_artifacts_workspace,
                target_workspace=target_workspace,
            )

        copy_collection_items, _ = self.work_request_ensure_child_server(
            task_name="copycollectionitems",
            task_data=CopyCollectionItemsData(copies=copies),
            # TODO: Include subject in workflow data.
            workflow_data=WorkRequestWorkflowData(
                display_name="Copy packages into target suite",
                step="copy-collection-items",
            ),
        )
        self.requires_artifact(copy_collection_items, copies[0].source_items)

        if self.data.update_indexes:
            update_suites_callback, _ = self.work_request_ensure_child_internal(
                task_name="workflow",
                # TODO: Include subject in workflow data.
                workflow_data=WorkRequestWorkflowData(
                    display_name="Start workflow to update repository indexes",
                    step="trigger-update-suites",
                ),
            )
            update_suites_callback.add_dependency(copy_collection_items)
            self.orchestrate_child(update_suites_callback)

    def callback_trigger_update_suites(self) -> bool:
        """
        Update suites once the items have been copied.

        We have to run this in the target workspace, which might not be the
        same as this workflow's workspace.  That involves triggering a
        separate workflow using a well-known template.
        """
        # TODO: At the moment, this only works for users who have the OWNER
        # role on the target workspace.
        # https://salsa.debian.org/freexian-team/debusine/-/issues/634 is
        # needed to fix this.
        template, _ = WorkflowTemplate.objects.get_or_create(
            name=WellKnownWorkflowTemplate.UPDATE_SUITES,
            task_name="update_suites",
            workspace=self.target_suite.collection.workspace,
            runtime_parameters=RuntimeParameter.ANY,
        )
        WorkRequest.objects.create_workflow(
            template=template,
            data=UpdateSuitesData(
                only_suites=[self.target_suite.collection.name]
            ).model_dump(exclude_unset=True),
        ).mark_pending()
        return True

    @override
    def compute_dynamic_data(self) -> PackagePublishWorkflowDynamicData:
        """
        Compute dynamic data for this workflow.

        :subject: source package names of ``binary_artifacts`` separated
          by spaces
        :parameter_summary: source package names of ``binary_artifacts``
          separated by spaces → suite
        :source_artifact_id: id of the source artifact
        :binary_artifacts_ids: ids of the binary artifacts
        :target_suite_id: collection id of the target_suite
        """
        binaries = self.work_request.lookup_multiple(
            self.data.binary_artifacts,
            expect_type=LookupChildType.ARTIFACT_OR_PROMISE,
        )

        source_package_names = " ".join(
            workflow_utils.get_source_package_names(
                binaries,
                configuration_key="binary_artifacts",
                artifact_expected_categories=(
                    ArtifactCategory.BINARY_PACKAGE,
                    ArtifactCategory.UPLOAD,
                ),
            )
        )

        try:
            source_artifact_id = workflow_utils.source_package(self).id
        except LookupError:
            source_artifact_id = None

        dynamic_data = PackagePublishWorkflowDynamicData(
            subject=source_package_names,
            parameter_summary=f"{source_package_names} → "
            f"{self.data.target_suite}",
            source_artifact_id=source_artifact_id,
            binary_artifacts_ids=artifact_ids(binaries),
            target_suite_id=self.target_suite.collection_id,
        )
        self.save_field_values(dynamic_data)
        return dynamic_data

    @override
    def get_input_artifacts(self) -> list[InputArtifact]:
        """Return input artifacts."""
        artifacts: list[InputArtifact] = []
        if self.data.source_artifact is not None:
            artifacts.append(
                InputArtifactSingle(
                    lookup=self.data.source_artifact,
                    label="source_artifact",
                    artifact_id=(
                        self.dynamic_data.source_artifact_id
                        if self.dynamic_data is not None
                        else None
                    ),
                )
            )

        artifacts.append(
            InputArtifactMultiple.from_ids(
                lookup=self.data.binary_artifacts,
                label="binary_artifacts",
                artifact_ids=(
                    self.dynamic_data.binary_artifacts_ids
                    if self.dynamic_data is not None
                    else None
                ),
            )
        )

        return artifacts



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views related to workers."""

from typing import Any, cast, override

from django.db.models import QuerySet
from django.shortcuts import get_object_or_404

from debusine.db.context import context
from debusine.db.models import WorkRequest, Worker
from debusine.db.models.workers import WorkerQuerySet
from debusine.tasks.models import WorkerType
from debusine.web.views.base import BaseUIView, DetailViewBase, ListViewBase
from debusine.web.views.places import Place
from debusine.web.views.table import TableMixin
from debusine.web.views.tables import WorkRequestTable, WorkerTable


class WorkersListView(BaseUIView, TableMixin[Worker], ListViewBase[Worker]):
    """List workers."""

    model = Worker
    base_template = "web/_base_debusine.html"
    template_name = "web/worker-list.html"
    title = "List of workers"
    table_class = WorkerTable
    paginate_by = 50

    @override
    def get_queryset(self) -> WorkerQuerySet[Any]:
        """Use the custom QuerySet."""
        queryset = super().get_queryset().exclude(worker_type=WorkerType.CELERY)
        queryset = cast(
            WorkerQuerySet[Any], queryset.select_related("worker_pool")
        )
        return queryset.active()

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        """Add context to the default ListView data."""
        ctx = super().get_context_data(**kwargs)
        ctx["WorkerType"] = {wt.name: wt.value for wt in WorkerType}
        return ctx


class WorkerDetailView(BaseUIView, DetailViewBase[Worker]):
    """Show details about a worker."""

    model = Worker
    base_template = "web/_base_debusine.html"
    template_name = "web/worker-detail.html"
    context_object_name = "worker"

    @override
    def get_object(
        self, queryset: QuerySet[Worker, Worker] | None = None
    ) -> Worker:
        """Return the Worker object to show."""
        assert queryset is None
        queryset = self.get_queryset().exclude(worker_type=WorkerType.CELERY)
        return get_object_or_404(queryset, name=self.kwargs["name"])

    @override
    def get_place(self) -> Place:
        return self.object.ui(self.request).place

    @override
    def get_context_data(self, **kwargs: Any) -> dict[str, Any]:
        ctx = super().get_context_data(**kwargs)
        ctx["show_metadata"] = self.request.user.is_authenticated
        ctx["work_requests"] = WorkRequestTable(
            self.request,
            cast(
                QuerySet[WorkRequest],
                self.object.assigned_work_requests.can_display(context.pc),
            ),
        ).get_paginator(per_page=10)
        return ctx



# Copyright © The Debusine Developers
# See the AUTHORS file at the top-level directory of this distribution
#
# This file is part of Debusine. It is subject to the license terms
# in the LICENSE file found in the top-level directory of this
# distribution. No part of Debusine, including this file, may be copied,
# modified, propagated, or distributed except according to the terms
# contained in the LICENSE file.

"""Views for the server application: collection relations."""

import hashlib
from typing import Any, override
from urllib.parse import urlencode

from django.db import IntegrityError, transaction
from django.db.models import QuerySet
from django.urls import reverse
from rest_framework import mixins, status
from rest_framework.parsers import JSONParser
from rest_framework.request import Request
from rest_framework.response import Response

from debusine.client.models import LookupChildType
from debusine.db.context import context
from debusine.db.models import Collection, CollectionRelation, User, Workspace
from debusine.db.permissioncontext import PermissionContext
from debusine.server.collections.lookup import lookup_single
from debusine.server.exceptions import DebusineAPIException
from debusine.server.serializers import (
    CollectionIdentifierSerializer,
    CollectionRelationSerializer,
    CollectionRelationSetSerializer,
)
from debusine.server.views.base import BaseAPIView, GenericViewSetBase
from debusine.server.views.rest import IsUserTokenAuthenticated


class CollectionRelationListView(
    BaseAPIView,
    mixins.ListModelMixin,
    GenericViewSetBase[CollectionRelation],
):
    """
    List relations, optionally filtered by source, target, and type.

    If filtered by source and type: return one page with all the relations
    and include the ETag header. Otherwise: paginates without the ETag header.
    """

    permission_classes = [IsUserTokenAuthenticated]
    parser_classes = [JSONParser]
    serializer_class = CollectionRelationSerializer

    _etag: str | None  # set only when forcing a single page

    @override
    def initial(self, request: Request, *args: Any, **kwargs: Any) -> None:
        """Set current workspace."""
        super().initial(request, *args, **kwargs)
        self.set_current_workspace(kwargs["workspace"])
        self._etag = None

    def _force_one_page(self) -> bool:
        """If "source" and "type" in query_params: return True."""
        query_params = self.request.GET
        return "source" in query_params and "type" in query_params

    @override
    def paginate_queryset[Row](
        self,
        queryset: QuerySet[CollectionRelation, Row],
    ) -> list[Row] | None:
        """Use one page when specifying "source" and "type" parameters."""
        if self._force_one_page():
            assert self.paginator is not None
            assert hasattr(self.paginator, "page_size")
            self.paginator.page_size = max(queryset.count(), 1)

        return super().paginate_queryset(queryset)

    @override
    def get_queryset(self) -> QuerySet[CollectionRelation]:
        """List collection relations in a workspace."""
        workspace = context.require_workspace()
        pc = context.pc
        assert isinstance(pc.user, User)

        self.enforce(workspace.can_display)

        query_params = self.request.GET

        source_collection = None
        target_collection = None

        if (source_query_param := query_params.get("source")) is not None:
            source_collection = _lookup_collection(
                source_query_param, workspace=workspace, pc=pc
            )
            self.enforce(source_collection.can_display)

        if (target_query_param := query_params.get("target")) is not None:
            target_collection = _lookup_collection(
                target_query_param, workspace=workspace, pc=pc
            )
            self.enforce(target_collection.can_display)

        queryset = _get_queryset(
            source_workspace=workspace,
            source=source_collection,
            target=target_collection,
            relation_type=query_params.get("type"),
        )

        if self._force_one_page():
            self._etag = _calculate_etag(queryset)

        return queryset

    @override
    def list(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        """List the relationships."""
        response = super().list(request, *args, **kwargs)

        if self._etag is not None:
            response["ETag"] = f'"{self._etag}"'

        return response


class CollectionRelationSetView(
    BaseAPIView,
):
    """Replace relation targets for one source collection and relation type."""

    permission_classes = [IsUserTokenAuthenticated]

    def put(self, request: Request, *args: Any, **kwargs: Any) -> Response:
        """Update set of relations of specific type from a collection."""
        self.set_current_workspace(kwargs["workspace"])
        workspace_model = context.require_workspace()
        self.enforce(workspace_model.can_display)

        pc = context.pc
        assert isinstance(pc.user, User)

        _validate_relation_type(kwargs["relation_type"])
        relation_type = CollectionRelation.Types(kwargs["relation_type"])

        source = _lookup_collection(
            kwargs["collection"], workspace=workspace_model, pc=pc
        )
        self.enforce(source.can_create_collection_relation)

        serializer = CollectionRelationSetSerializer(data=request.data)
        serializer.is_valid(raise_exception=True)

        with transaction.atomic():
            target_models = []

            for target in serializer.validated_data["targets"]:
                try:
                    target_models.append(
                        workspace_model.get_collection(
                            pc=pc,
                            name=target["name"],
                            category=target["category"],
                        )
                    )
                except Collection.DoesNotExist:
                    collection_lookup = f"{target["name"]}@{target["category"]}"
                    raise DebusineAPIException(
                        title=(
                            f'Collection {collection_lookup!r} does not exist '
                            'or is not accessible'
                        ),
                        status_code=status.HTTP_400_BAD_REQUEST,
                    )

            source = Collection.objects.select_for_update().get(pk=source.pk)

            current_etag = _calculate_etag(
                _get_queryset(
                    source_workspace=source.workspace,
                    source=source,
                    relation_type=relation_type,
                )
            )

            list_url = (
                reverse(
                    "api:collection-relation-list",
                    kwargs={"workspace": workspace_model.name},
                )
                + "?"
                + urlencode(
                    {
                        "source": f"{source.name}@{source.category}",
                        "type": relation_type,
                    }
                )
            )

            if_match = self.request.headers.get("If-Match")

            if if_match is None:
                raise DebusineAPIException(
                    title='"If-Match" header is required',
                    detail=(
                        'The "If-Match" header is required for PUT. '
                        f'Fetch the current ETag from {list_url} and retry.'
                    ),
                    status_code=status.HTTP_400_BAD_REQUEST,
                )

            if f'"{current_etag}"' != if_match:
                raise DebusineAPIException(
                    title="Collection relations changed",
                    detail=(
                        'The "If-Match" header does not match the '
                        f'current ETag. Fetch the current ETag from '
                        f'"{list_url}" and retry.'
                    ),
                    status_code=status.HTTP_412_PRECONDITION_FAILED,
                )

            CollectionRelation.objects.filter(
                source=source, type=relation_type
            ).delete()

            try:
                response_targets = []

                for position, target_model in enumerate(target_models, start=1):
                    collection_relation_position = (
                        position
                        if relation_type
                        == CollectionRelation.Types.SUITE_REQUIRES
                        else None
                    )
                    collection_relation = CollectionRelation.objects.create(
                        source=source,
                        target=target_model,
                        type=relation_type,
                        position=collection_relation_position,
                        created_by_user=pc.user,
                    )
                    response_targets.append(
                        CollectionIdentifierSerializer(
                            {
                                "name": collection_relation.target.name,
                                "category": collection_relation.target.category,
                            }
                        ).data
                    )
            except IntegrityError as exc:
                raise DebusineAPIException(
                    title="Invalid relation targets",
                    detail=str(exc),
                    status_code=status.HTTP_400_BAD_REQUEST,
                ) from exc

        serializer = CollectionRelationSetSerializer(
            {"targets": response_targets}
        )

        return Response(
            serializer.data,
            status=status.HTTP_200_OK,
        )


def _lookup_collection(
    collection_identifier: str,
    *,
    workspace: Workspace,
    pc: PermissionContext,
) -> Collection:
    if "/" in collection_identifier:
        raise DebusineAPIException(
            title="Invalid collection identifier in query parameter",
            detail=(
                f"Collection identifier {collection_identifier!r} "
                "must not contain '/'"
            ),
            status_code=status.HTTP_400_BAD_REQUEST,
        )

    try:
        result = lookup_single(
            collection_identifier,
            workspace,
            pc=pc,
            expect_type=LookupChildType.COLLECTION,
        )
    except KeyError:
        raise DebusineAPIException(
            title="Collection not found",
            detail=(
                f"Collection {collection_identifier!r} does not exist "
                f"in workspace {workspace.name!r}"
            ),
            status_code=status.HTTP_404_NOT_FOUND,
        )
    except LookupError as exc:
        raise DebusineAPIException(
            title="Invalid collection identifier in query parameter",
            detail=str(exc),
            status_code=status.HTTP_400_BAD_REQUEST,
        ) from exc

    return result.collection


def _validate_relation_type(relation_type: str) -> None:
    if relation_type not in CollectionRelation.Types.values:
        raise DebusineAPIException(
            title="Invalid relation type",
            detail=f"Unknown relation type: {relation_type!r}",
            status_code=status.HTTP_400_BAD_REQUEST,
        )


def _calculate_etag(collection_relations: QuerySet[CollectionRelation]) -> str:
    """Return a stable ETag for the ordered relation set."""
    h = hashlib.sha256()

    for cr in collection_relations:
        h.update(
            f"{cr.source.id}|{cr.target.id}|{cr.type}|{cr.position}\n".encode()
        )

    return h.hexdigest()


def _get_queryset(
    *,
    source_workspace: Workspace,
    source: Collection | None = None,
    relation_type: str | None = None,
    target: Collection | None = None,
) -> QuerySet[CollectionRelation]:
    queryset = CollectionRelation.objects.filter(
        source__workspace=source_workspace
    ).select_related("source", "target")

    if source is not None:
        queryset = queryset.filter(source=source)

    if target is not None:
        queryset = queryset.filter(target=target)

    if relation_type is not None:
        _validate_relation_type(relation_type)
        queryset = queryset.filter(type=relation_type)

    queryset = queryset.order_by(
        "source__category",
        "source__name",
        "type",
        "position",
        "target__category",
        "target__name",
        "id",
    )

    return queryset



# Generated by Django 4.2.23 on 2025-12-12 12:14

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ('db', '0039_alter_worker_worker_type_and_more'),
    ]

    operations = [
        migrations.AddField(
            model_name='workrequest',
            name='pending_at',
            field=models.DateTimeField(blank=True, null=True),
        ),
    ]
